#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Mask related classes."""

from typing import Any, Dict, Optional, Type, TypeVar

from tensorbay.label.basic import AttributeType, SubcatalogBase
from tensorbay.label.supports import AttributesMixin, IsTrackingMixin, MaskCategoriesMixin
from tensorbay.utility import URL, FileMixin, RemoteFileMixin, ReprMixin


class SemanticMaskSubcatalog(SubcatalogBase, MaskCategoriesMixin, AttributesMixin):
    """This class defines the subcatalog for semantic mask type of labels.

    Attributes:
        description: The description of the entire semantic mask subcatalog.
        categories: All the possible categories in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the category names as keys
            and the :class:`~tensorbay.label.supports.CategoryInfo` as values.
        category_delimiter: The delimiter in category values indicating parent-child relationship.
        attributes: All the possible attributes in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the attribute names as keys
            and the :class:`~tensorbay.label.attribute.AttributeInfo` as values.
        is_tracking: Whether the Subcatalog contains tracking information.

    Examples:
        *Initialization Method 1:* Init from ``SemanticMaskSubcatalog.loads()`` method.

        >>> catalog = {
        ...     "SEMANTIC_MASK": {
        ...         "categories": [
        ...             {'name': 'cat', "categoryId": 1},
        ...             {'name': 'dog', "categoryId": 2}
        ...         ],
        ...         "attributes": [{'name': 'occluded', 'type': 'boolean'}],
        ...     }
        ... }
        >>> SemanticMaskSubcatalog.loads(catalog["SEMANTIC_MASK"])
        SemanticMaskSubcatalog(
          (categories): NameList [...],
          (attributes): NameList [...]
        )

        *Initialization Method 2:* Init an empty SemanticMaskSubcatalog and then add the attributes.

        >>> semantic_mask_subcatalog = SemanticMaskSubcatalog()
        >>> semantic_mask_subcatalog.add_category("cat", 1)
        >>> semantic_mask_subcatalog.add_category("dog", 2)
        >>> semantic_mask_subcatalog.add_attribute("occluded", type_="boolean")
        >>> semantic_mask_subcatalog
        SemanticMaskSubcatalog(
          (categories): NameList [...],
          (attributes): NameList [...]
        )

    """


class InstanceMaskSubcatalog(SubcatalogBase, MaskCategoriesMixin, IsTrackingMixin, AttributesMixin):
    """This class defines the subcatalog for instance mask type of labels.

    Attributes:
        description: The description of the entire instance mask subcatalog.
        categories: All the possible categories in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the category names as keys
            and the :class:`~tensorbay.label.supports.CategoryInfo` as values.
        category_delimiter: The delimiter in category values indicating parent-child relationship.
        attributes: All the possible attributes in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the attribute names as keys
            and the :class:`~tensorbay.label.attribute.AttributeInfo` as values.
        is_tracking: Whether the Subcatalog contains tracking information.

    Examples:
        *Initialization Method 1:* Init from ``InstanceMaskSubcatalog.loads()`` method.

        >>> catalog = {
        ...     "INSTANCE_MASK": {
        ...         "categories": [
        ...             {'name': 'background', "categoryId": 0}
        ...         ],
        ...         "attributes": [{'name': 'occluded', 'type': 'boolean'}],
        ...     }
        ... }
        >>> InstanceMaskSubcatalog.loads(catalog["INSTANCE_MASK"])
        InstanceMaskSubcatalog(
          (is_tracking): False,
          (categories): NameList [...],
          (attributes): NameList [...]
        )

        *Initialization Method 2:* Init an empty InstanceMaskSubcatalog and then add the attributes.

        >>> instance_mask_subcatalog = InstanceMaskSubcatalog()
        >>> instance_mask_subcatalog.add_category("background", 0)
        >>> instance_mask_subcatalog.add_attribute("occluded", type_="boolean")
        >>> instance_mask_subcatalog
        InstanceMaskSubcatalog(
          (categories): NameList [...],
          (attributes): NameList [...]
        )

    """


class PanopticMaskSubcatalog(SubcatalogBase, MaskCategoriesMixin, AttributesMixin):
    """This class defines the subcatalog for panoptic mask type of labels.

    Attributes:
        description: The description of the entire panoptic mask subcatalog.
        categories: All the possible categories in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the category names as keys
            and the :class:`~tensorbay.label.supports.CategoryInfo` as values.
        category_delimiter: The delimiter in category values indicating parent-child relationship.
        attributes: All the possible attributes in the corresponding dataset
            stored in a :class:`~tensorbay.utility.name.NameList`
            with the attribute names as keys
            and the :class:`~tensorbay.label.attribute.AttributeInfo` as values.
        is_tracking: Whether the Subcatalog contains tracking information.

    Examples:
        *Initialization Method 1:* Init from ``PanopticMaskSubcatalog.loads()`` method.

        >>> catalog = {
        ...     "PANOPTIC_MASK": {
        ...         "categories": [
        ...             {'name': 'cat', "categoryId": 1},
        ...             {'name': 'dog', "categoryId": 2}
        ...         ],
        ...         "attributes": [{'name': 'occluded', 'type': 'boolean'}],
        ...     }
        ... }
        >>> PanopticMaskSubcatalog.loads(catalog["PANOPTIC_MASK"])
        PanopticMaskSubcatalog(
          (categories): NameList [...],
          (attributes): NameList [...]
        )

        *Initialization Method 2:* Init an empty PanopticMaskSubcatalog and then add the attributes.

        >>> panoptic_mask_subcatalog = PanopticMaskSubcatalog()
        >>> panoptic_mask_subcatalog.add_category("cat", 1)
        >>> panoptic_mask_subcatalog.add_category("dog", 2)
        >>> panoptic_mask_subcatalog.add_attribute("occluded", type_="boolean")
        >>> panoptic_mask_subcatalog
        PanopticMaskSubcatalog(
          (categories): NameList [...],
          (attributes): NameList [...]
        )

    """


class SemanticMaskBase(ReprMixin):
    """SemanticMaskBase is a base class for the semantic mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the category id,
            and the value is the corresponding attributes.

    """

    _repr_attrs = ("all_attributes",)

    all_attributes: Dict[int, AttributeType]


class InstanceMaskBase(ReprMixin):
    """InstanceMaskBase is a base class for the instance mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the instance id,
            and the value is the corresponding attributes.

    """

    _repr_attrs = ("all_attributes",)

    all_attributes: Dict[int, AttributeType]


class PanopticMaskBase(ReprMixin):
    """PanopticMaskBase is a base class for the panoptic mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the instance id,
            and the value is the corresponding attributes.
        all_category_ids: The dict of the category id in this mask, which key is the instance id,
            and the value is the corresponding category id.

    """

    _repr_attrs = ("all_category_ids", "all_attributes")

    all_attributes: Dict[int, AttributeType]

    def __init__(self) -> None:
        self.all_category_ids: Dict[int, int] = {}


class SemanticMask(SemanticMaskBase, FileMixin):
    """SemanticMask is a class for the local semantic mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the category id,
            and the value is the corresponding attributes.

    """

    def __init__(self, local_path: str) -> None:
        FileMixin.__init__(self, local_path)

    def get_callback_body(self) -> Dict[str, Any]:
        """Get the callback request body for uploading.

        Returns:
            The callback request body, which looks like::

                    {
                        "checksum": <str>,
                        "fileSize": <int>,
                        "info": [
                            {
                                "categoryId": 0,
                                "attributes": {
                                    "occluded": True
                                }
                            },
                            {
                                "categoryId": 1,
                                "attributes": {
                                    "occluded": False
                                }
                            }
                        ]
                    }

        """
        body = super()._get_callback_body()
        if hasattr(self, "all_attributes"):
            body["info"] = [
                {"categoryId": i, "attributes": attributes}
                for i, attributes in self.all_attributes.items()
            ]
        return body


class InstanceMask(InstanceMaskBase, FileMixin):
    """InstanceMask is a class for the local instance mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the instance id,
            and the value is the corresponding attributes.

    """

    def __init__(self, local_path: str) -> None:
        FileMixin.__init__(self, local_path)

    def get_callback_body(self) -> Dict[str, Any]:
        """Get the callback request body for uploading.

        Returns:
            The callback request body, which looks like::

                    {
                        "checksum": <str>,
                        "fileSize": <int>,
                        "info": [
                            {
                                "instanceId": 0,
                                "attributes": {
                                    "occluded": True
                                }
                            },
                            {
                                "instanceId": 1,
                                "attributes": {
                                    "occluded": False
                                }
                            }
                        ]
                    }

        """
        body = super()._get_callback_body()
        if hasattr(self, "all_attributes"):
            body["info"] = [
                {"instanceId": i, "attributes": attributes}
                for i, attributes in self.all_attributes.items()
            ]
        return body


class PanopticMask(PanopticMaskBase, FileMixin):
    """PanopticMask is a class for the local panoptic mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the instance id,
            and the value is the corresponding attributes.
        all_category_ids: The dict of the category id in this mask, which key is the instance id,
            and the value is the corresponding category id.

    """

    def __init__(self, local_path: str) -> None:
        PanopticMaskBase.__init__(self)
        FileMixin.__init__(self, local_path)

    def get_callback_body(self) -> Dict[str, Any]:
        """Get the callback request body for uploading.

        Returns:
            The callback request body, which looks like::

                    {
                        "checksum": <str>,
                        "fileSize": <int>,
                        "info": [
                            {
                                "instanceId": 0,
                                "categoryId": 100,
                                "attributes": {
                                    "occluded": True
                                }
                            },
                            {
                                "instanceId": 1,
                                "categoryId": 101,
                                "attributes": {
                                    "occluded": False
                                }
                            }
                        ]
                    }

        """
        body = super()._get_callback_body()
        all_attributes = getattr(self, "all_attributes", None)
        info = []
        for i, category_id in self.all_category_ids.items():
            item = {"instanceId": i, "categoryId": category_id}
            if all_attributes:
                item["attributes"] = all_attributes[i]  # pylint: disable=unsubscriptable-object
            info.append(item)

        body["info"] = info
        return body


class RemoteSemanticMask(SemanticMaskBase, RemoteFileMixin):
    """RemoteSemanticMask is a class for the remote semantic mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the category id,
            and the value is the corresponding attributes.

    """

    _T = TypeVar("_T", bound="RemoteSemanticMask")

    @classmethod
    def from_response_body(cls: Type[_T], body: Dict[str, Any]) -> _T:
        """Loads a :class:`RemoteSemanticMask` object from a response body.

        Arguments:
            body: The response body which contains the information of a remote semantic mask,
                whose format should be like::

                    {
                        "remotePath": <str>,
                        "info": [
                            {
                                "categoryId": 0,
                                "attributes": {
                                    "occluded": True
                                }
                            },
                            {
                                "categoryId": 1,
                                "attributes": {
                                    "occluded": False
                                }
                            }
                        ]
                    }

        Returns:
            The loaded :class:`RemoteSemanticMask` object.

        """
        mask = cls(body["remotePath"])
        if "info" in body:
            mask.all_attributes = {item["categoryId"]: item["attributes"] for item in body["info"]}

        return mask


class RemoteInstanceMask(InstanceMaskBase, RemoteFileMixin):
    """RemoteInstanceMask is a class for the remote instance mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the instance id,
            and the value is the corresponding attributes.

    """

    _T = TypeVar("_T", bound="RemoteInstanceMask")

    @classmethod
    def from_response_body(cls: Type[_T], body: Dict[str, Any]) -> _T:
        """Loads a :class:`RemoteInstanceMask` object from a response body.

        Arguments:
            body: The response body which contains the information of a remote instance mask,
                whose format should be like::

                    {
                        "remotePath": <str>,
                        "info": [
                            {
                                "instanceId": 0,
                                "attributes": {
                                    "occluded": True
                                }
                            },
                            {
                                "instanceId": 1,
                                "attributes": {
                                    "occluded": False
                                }
                            }
                        ]
                    }

        Returns:
            The loaded :class:`RemoteInstanceMask` object.

        """
        mask = cls(body["remotePath"])
        if "info" in body:
            mask.all_attributes = {item["instanceId"]: item["attributes"] for item in body["info"]}

        return mask


class RemotePanopticMask(PanopticMaskBase, RemoteFileMixin):
    """RemotePanoticMask is a class for the remote panotic mask label.

    Attributes:
        all_attributes: The dict of the attributes in this mask, which key is the instance id,
            and the value is the corresponding attributes.

    """

    _T = TypeVar("_T", bound="RemotePanopticMask")

    def __init__(self, remote_path: str, *, url: Optional[URL] = None) -> None:
        PanopticMaskBase.__init__(self)
        RemoteFileMixin.__init__(self, remote_path, url=url)

    @classmethod
    def from_response_body(cls: Type[_T], body: Dict[str, Any]) -> _T:
        """Loads a :class:`RemotePanopticMask` object from a response body.

        Arguments:
            body: The response body which contains the information of a remote panoptic mask,
                whose format should be like::

                    {
                        "remotePath": <str>,
                        "info": [
                            {
                                "instanceId": 0,
                                "categoryId": 100,
                                "attributes": {
                                    "occluded": True
                                }
                            },
                            {
                                "instanceId": 1,
                                "categoryId": 101,
                                "attributes": {
                                    "occluded": False
                                }
                            }
                        ]
                    }

        Returns:
            The loaded :class:`RemotePanopticMask` object.

        """
        mask = cls(body["remotePath"])
        info = body["info"]
        # Workaround for panoptic masks with empty info.
        # TODO: Remove this check after reuploading COCO mask labels or changing standards.
        if info:
            mask.all_category_ids = {item["instanceId"]: item["categoryId"] for item in info}
            if "attributes" in info[0]:
                mask.all_attributes = {item["instanceId"]: item["attributes"] for item in info}

        return mask



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class GAS.

The :class:`GAS` defines the initial client to interact between local and TensorBay.
It provides some operations on datasets level such as :meth:`GAS.create_dataset`,
:meth:`GAS.list_dataset_names` and :meth:`GAS.get_dataset`.

AccessKey is required when operating with dataset.

"""

import logging
from typing import Any, Dict, Generator, Optional, Type, Union, overload

from typing_extensions import Literal

from tensorbay.client.cloud_storage import CloudClient, StorageConfig
from tensorbay.client.dataset import DatasetClient, FusionDatasetClient
from tensorbay.client.lazy import PagingList
from tensorbay.client.log import UPLOAD_DATASET_RESUME_TEMPLATE
from tensorbay.client.requests import Client
from tensorbay.client.status import Status
from tensorbay.client.struct import ROOT_COMMIT_ID, UserInfo
from tensorbay.dataset import Dataset, FusionDataset
from tensorbay.exception import DatasetTypeError, ResourceNotExistError
from tensorbay.utility import Tqdm

DatasetClientType = Union[DatasetClient, FusionDatasetClient]

logger = logging.getLogger(__name__)

DEFAULT_BRANCH = "main"
DEFAULT_IS_PUBLIC = False


class GAS:
    """:class:`GAS` defines the initial client to interact between local and TensorBay.

    :class:`GAS` provides some operations on dataset level such as
    :meth:`GAS.create_dataset` :meth:`GAS.list_dataset_names` and :meth:`GAS.get_dataset`.

    Arguments:
        access_key: User's access key.
        url: The host URL of the gas website.

    """

    def __init__(self, access_key: str, url: str = "") -> None:
        self._client = Client(access_key, url)

    def _generate_auth_storage_configs(
        self, name: Optional[str] = None, offset: int = 0, limit: int = 128
    ) -> Generator[StorageConfig, None, int]:
        params: Dict[str, Any] = {"offset": offset, "limit": limit}
        if name:
            params["name"] = name

        response = self._client.open_api_do("GET", "storage-configs", "", params=params).json()

        for config in response["configs"]:
            yield StorageConfig.loads(config)

        return response["totalCount"]  # type: ignore[no-any-return]

    def _get_dataset_with_any_type(self, name: str) -> DatasetClientType:
        info = self._get_dataset(name)
        dataset_id = info["id"]
        is_fusion = info["type"]
        commit_id = info["commitId"]
        default_branch = info["defaultBranch"]
        dataset_alias = info["alias"]
        is_public = info["isPublic"]

        status = Status(default_branch, commit_id=commit_id)

        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(
            name,
            dataset_id,
            self,
            status=status,
            alias=dataset_alias,
            is_public=is_public,
        )

    def _get_dataset(self, name: str) -> Dict[str, Any]:
        """Get the information of the TensorBay dataset with the input name.

        Arguments:
            name: The name of the requested dataset.

        Returns:
            The dict of dataset information.

        Raises:
            ResourceNotExistError: When the required dataset does not exist.

        """
        if not name:
            raise ResourceNotExistError(resource="dataset", identification=name)

        try:
            response = self._list_datasets(name=name)
            info = response["datasets"][0]
        except IndexError as error:
            raise ResourceNotExistError(resource="dataset", identification=name) from error

        dataset_id = info["id"]
        response = self._client.open_api_do("GET", "", dataset_id).json()
        response["id"] = dataset_id
        return response

    def _list_datasets(
        self,
        name: Optional[str] = None,
        offset: int = 0,
        limit: int = 128,
    ) -> Dict[str, Any]:
        params: Dict[str, Any] = {
            "offset": offset,
            "limit": limit,
        }
        if name:
            params["name"] = name

        response = self._client.open_api_do("GET", "", params=params)
        return response.json()  # type: ignore[no-any-return]

    def _generate_dataset_names(
        self,
        name: Optional[str] = None,
        offset: int = 0,
        limit: int = 128,
    ) -> Generator[str, None, int]:
        response = self._list_datasets(name, offset, limit)

        for item in response["datasets"]:
            yield item["name"]

        return response["totalCount"]  # type: ignore[no-any-return]

    def get_user(self) -> UserInfo:
        """Get the user information with the current accesskey.

        Returns:
            The :class:`.struct.UserInfo` with the current accesskey.

        """
        response = self._client.open_api_do("GET", "users").json()
        return UserInfo.loads(response)

    def get_auth_storage_config(self, name: str) -> StorageConfig:
        """Get the auth storage config with the given name.

        Arguments:
            name: The required auth storage config name.

        Returns:
            The auth storage config with the given name.

        Raises:
            TypeError: When the given auth storage config is illegal.
            ResourceNotExistError: When the required auth storage config does not exist.

        """
        if not name:
            raise TypeError("The given auth storage config name is illegal")

        try:
            config = next(self._generate_auth_storage_configs(name))
        except StopIteration as error:
            raise ResourceNotExistError(
                resource="auth storage config", identification=name
            ) from error

        return config

    def list_auth_storage_configs(self) -> PagingList[StorageConfig]:
        """List auth storage configs.

        Returns:
            The PagingList of all auth storage configs.

        """
        return PagingList(
            lambda offset, limit: self._generate_auth_storage_configs(None, offset, limit),
            128,
        )

    def delete_storage_config(self, name: str) -> None:
        """Delete a storage config in TensorBay.

        Arguments:
            name: Name of the storage config, unique for a team.

        """
        self._client.open_api_do("DELETE", f"storage-configs/{name}")

    def create_oss_storage_config(
        self,
        name: str,
        file_path: str,
        *,
        endpoint: str,
        accesskey_id: str,
        accesskey_secret: str,
        bucket_name: str,
    ) -> CloudClient:
        """Create an oss auth storage config.

        Arguments:
            name: The required auth storage config name.
            file_path: dataset storage path of the bucket.
            endpoint: endpoint of the oss.
            accesskey_id: accesskey_id of the oss.
            accesskey_secret: accesskey_secret of the oss.
            bucket_name: bucket_name of the oss.

        Returns:
            The cloud client of this dataset.

        """
        post_data = {
            "name": name,
            "filePath": file_path,
            "endpoint": endpoint,
            "accesskeyId": accesskey_id,
            "accesskeySecret": accesskey_secret,
            "bucketName": bucket_name,
        }

        self._client.open_api_do("POST", "storage-configs/oss", json=post_data)
        return CloudClient(name, self._client)

    def create_s3_storage_config(
        self,
        name: str,
        file_path: str,
        *,
        endpoint: str,
        accesskey_id: str,
        accesskey_secret: str,
        bucket_name: str,
    ) -> CloudClient:
        """Create a s3 auth storage config.

        Arguments:
            name: The required auth storage config name.
            file_path: dataset storage path of the bucket.
            endpoint: endpoint of the s3.
            accesskey_id: accesskey_id of the s3.
            accesskey_secret: accesskey_secret of the s3.
            bucket_name: bucket_name of the s3.

        Returns:
            The cloud client of this dataset.

        """
        post_data = {
            "name": name,
            "filePath": file_path,
            "endpoint": endpoint,
            "accesskeyId": accesskey_id,
            "accesskeySecret": accesskey_secret,
            "bucketName": bucket_name,
        }

        self._client.open_api_do("POST", "storage-configs/s3", json=post_data)
        return CloudClient(name, self._client)

    def create_azure_storage_config(
        self,
        name: str,
        file_path: str,
        *,
        account_type: str,
        account_name: str,
        account_key: str,
        container_name: str,
    ) -> CloudClient:
        """Create an azure auth storage config.

        Arguments:
            name: The required auth storage config name.
            file_path: dataset storage path of the bucket.
            account_type: account type of azure, only support "China" and "Global".
            account_name: account name of the azure.
            account_key: account key of the azure.
            container_name: container name of the azure.

        Returns:
            The cloud client of this dataset.

        """
        post_data = {
            "name": name,
            "filePath": file_path,
            "accesskeyId": account_name,
            "accesskeySecret": account_key,
            "containerName": container_name,
            "accountType": account_type,
        }

        self._client.open_api_do("POST", "storage-configs/azure", json=post_data)
        return CloudClient(name, self._client)

    def create_local_storage_config(self, name: str, file_path: str, endpoint: str) -> None:
        """Create a local auth storage config.

        Arguments:
            name: The required local storage config name.
            file_path: The dataset storage path under the local storage.
            endpoint: The external IP address of the local storage.

        """
        post_data = {
            "name": name,
            "filePath": file_path,
            "endpoint": endpoint,
        }

        self._client.open_api_do("POST", "storage-configs/local", json=post_data)

    def get_cloud_client(self, name: str) -> CloudClient:
        """Get a cloud client used for interacting with cloud platform.

        Arguments:
            name: The required auth storage config name.

        Returns:
            The cloud client of this dataset.

        """
        self.get_auth_storage_config(name)
        return CloudClient(name, self._client)

    @overload
    def create_dataset(
        self,
        name: str,
        is_fusion: Literal[False] = False,
        *,
        config_name: Optional[str] = None,
        alias: str = "",
        is_public: bool = DEFAULT_IS_PUBLIC,
    ) -> DatasetClient:
        ...

    @overload
    def create_dataset(
        self,
        name: str,
        is_fusion: Literal[True],
        *,
        config_name: Optional[str] = None,
        alias: str = "",
        is_public: bool = DEFAULT_IS_PUBLIC,
    ) -> FusionDatasetClient:
        ...

    @overload
    def create_dataset(
        self,
        name: str,
        is_fusion: bool = False,
        *,
        config_name: Optional[str] = None,
        alias: str = "",
        is_public: bool = DEFAULT_IS_PUBLIC,
    ) -> DatasetClientType:
        ...

    def create_dataset(
        self,
        name: str,
        is_fusion: bool = False,
        *,
        config_name: Optional[str] = None,
        alias: str = "",
        is_public: bool = DEFAULT_IS_PUBLIC,
    ) -> DatasetClientType:
        """Create a TensorBay dataset with given name.

        Arguments:
            name: Name of the dataset, unique for a user.
            is_fusion: Whether the dataset is a fusion dataset, True for fusion dataset.
            config_name: The auth storage config name.
            alias: Alias of the dataset, default is "".
            is_public: Whether the dataset is a public dataset.

        Returns:
            The created :class:`~tensorbay.client.dataset.DatasetClient` instance or
            :class:`~tensorbay.client.dataset.FusionDatasetClient` instance (is_fusion=True),
            and the status of dataset client is "commit".

        """
        post_data = {
            "name": name,
            "type": int(is_fusion),  # normal dataset: 0, fusion dataset: 1
            "alias": alias,
            "isPublic": is_public,
        }
        if config_name is not None:
            post_data["configName"] = config_name

        status = Status(DEFAULT_BRANCH, commit_id=ROOT_COMMIT_ID)

        response = self._client.open_api_do("POST", "", json=post_data)
        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(
            name,
            response.json()["id"],
            self,
            status=status,
            alias=alias,
            is_public=is_public,
        )

    @overload
    def get_dataset(self, name: str, is_fusion: Literal[False] = False) -> DatasetClient:
        ...

    @overload
    def get_dataset(self, name: str, is_fusion: Literal[True]) -> FusionDatasetClient:
        ...

    @overload
    def get_dataset(self, name: str, is_fusion: bool = False) -> DatasetClientType:
        ...

    def get_dataset(self, name: str, is_fusion: bool = False) -> DatasetClientType:
        """Get a TensorBay dataset with given name and commit ID.

        Arguments:
            name: The name of the requested dataset.
            is_fusion: Whether the dataset is a fusion dataset, True for fusion dataset.

        Returns:
            The requested :class:`~tensorbay.client.dataset.DatasetClient` instance or
            :class:`~tensorbay.client.dataset.FusionDatasetClient` instance (is_fusion=True),
            and the status of dataset client is "commit".

        Raises:
            DatasetTypeError: When the requested dataset type is not the same as given.

        """
        info = self._get_dataset(name)

        dataset_id = info["id"]
        type_flag = info["type"]
        commit_id = info["commitId"]
        default_branch = info["defaultBranch"]
        dataset_alias = info["alias"]
        is_public = info["isPublic"]

        status = Status(default_branch, commit_id=commit_id)

        if is_fusion != type_flag:
            raise DatasetTypeError(dataset_name=name, is_fusion=type_flag)
        ReturnType: Type[DatasetClientType] = FusionDatasetClient if is_fusion else DatasetClient
        return ReturnType(
            name,
            dataset_id,
            self,
            status=status,
            alias=dataset_alias,
            is_public=is_public,
        )

    def list_dataset_names(self) -> PagingList[str]:
        """List names of all TensorBay datasets.

        Returns:
            The PagingList of all TensorBay dataset names.

        """
        return PagingList(
            lambda offset, limit: self._generate_dataset_names(None, offset, limit), 128
        )

    def update_dataset(
        self,
        name: str,
        *,
        alias: Optional[str] = None,
        is_public: Optional[bool] = None,
    ) -> None:
        """Update a TensorBay Dataset.

        Arguments:
            name: Name of the dataset, unique for a user.
            alias: New alias of the dataset.
            is_public: Whether the dataset is public.

        """
        dataset_id = self._get_dataset(name)["id"]
        patch_data: Dict[str, Any] = {}
        if alias is not None:
            patch_data["alias"] = alias

        if is_public is not None:
            patch_data["isPublic"] = is_public

        if patch_data:
            self._client.open_api_do("PATCH", "", dataset_id, json=patch_data)

    def rename_dataset(self, name: str, new_name: str) -> None:
        """Rename a TensorBay Dataset with given name.

        Arguments:
            name: Name of the dataset, unique for a user.
            new_name: New name of the dataset, unique for a user.

        """
        dataset_id = self._get_dataset(name)["id"]
        patch_data: Dict[str, str] = {"name": new_name}
        self._client.open_api_do("PATCH", "", dataset_id, json=patch_data)

    @overload
    def upload_dataset(
        self,
        dataset: Dataset,
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> DatasetClient:
        ...

    @overload
    def upload_dataset(
        self,
        dataset: FusionDataset,
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> FusionDatasetClient:
        ...

    @overload
    def upload_dataset(
        self,
        dataset: Union[Dataset, FusionDataset],
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> DatasetClientType:
        ...

    def upload_dataset(
        self,
        dataset: Union[Dataset, FusionDataset],
        draft_number: Optional[int] = None,
        *,
        branch_name: Optional[str] = None,
        jobs: int = 1,
        skip_uploaded_files: bool = False,
        quiet: bool = False,
    ) -> DatasetClientType:
        """Upload a local dataset to TensorBay.

        This function will upload all information contains
        in the :class:`~tensorbay.dataset.dataset.Dataset`
        or :class:`~tensorbay.dataset.dataset.FusionDataset`, which includes:

            - Create a TensorBay dataset with the name and type of input local dataset.
            - Upload all :class:`~tensorbay.dataset.segment.Segment`
              or :class:`~tensorbay.dataset.segment.FusionSegment` in the dataset to TensorBay.

        Arguments:
            dataset: The :class:`~tensorbay.dataset.dataset.Dataset` or
                :class:`~tensorbay.dataset.dataset. FusionDataset` needs to be uploaded.
            draft_number: The draft number.
            branch_name: The branch name.
            jobs: The number of the max workers in multi-thread upload.
            skip_uploaded_files: Set it to True to skip the uploaded files.
            quiet: Set to True to stop showing the upload process bar.

        Returns:
            The :class:`~tensorbay.client.dataset.DatasetClient` or
            :class:`~tensorbay.client.dataset.FusionDatasetClient`
            bound with the uploaded dataset.

        Raises:
            ValueError: When uploading the dataset based on both draft number
                and branch name is not allowed.
            Exception: When Exception was raised during uploading dataset.

        """
        dataset_client = self.get_dataset(dataset.name, isinstance(dataset, FusionDataset))

        if draft_number and branch_name:
            raise ValueError(
                "Uploading the dataset based on both draft number and branch name is not allowed"
            )

        if draft_number:
            dataset_client.checkout(draft_number=draft_number)

        else:
            target_branch_name = branch_name if branch_name else dataset_client.status.branch_name

            drafts = dataset_client.list_drafts(branch_name=target_branch_name)
            if drafts:
                dataset_client.checkout(draft_number=drafts[0].number)
            else:
                dataset_client.create_draft(
                    'Draft autogenerated by "GAS.upload_dataset"', branch_name=target_branch_name
                )

        try:
            if dataset.catalog:
                dataset_client.upload_catalog(dataset.catalog)

            dataset_client.update_notes(**dataset.notes)  # type: ignore[arg-type]

            if isinstance(dataset, Dataset):
                data_count = sum(len(segment) for segment in dataset)
            else:
                data_count = sum(
                    sum(len(frame) for frame in fusion_segment) for fusion_segment in dataset
                )

            with Tqdm(data_count, disable=quiet) as pbar:
                for segment in dataset:
                    dataset_client._upload_segment(  # pylint: disable=protected-access
                        segment,  # type: ignore[arg-type]
                        jobs=jobs,
                        skip_uploaded_files=skip_uploaded_files,
                        pbar=pbar,
                    )
        except Exception:
            if draft_number:
                logger.error(
                    UPLOAD_DATASET_RESUME_TEMPLATE,
                    dataset_client.status.draft_number,
                    dataset_client.status.draft_number,
                )
            raise

        return dataset_client

    def delete_dataset(self, name: str) -> None:
        """Delete a TensorBay dataset with given name.

        Arguments:
            name: Name of the dataset, unique for a user.

        """
        dataset_id = self._get_dataset(name)["id"]
        self._client.open_api_do("DELETE", "", dataset_id)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of DogsVsCats dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME = "DogsVsCats"
_SEGMENTS = {"train": True, "test": False}


def DogsVsCats(path: str) -> Dataset:
    """`Dogs vs Cats <https://www.kaggle.com/c/dogs-vs-cats>`_ dataset.

    The file structure should be like::

        <path>
            train/
                cat.0.jpg
                ...
                dog.0.jpg
                ...
            test/
                1000.jpg
                1001.jpg
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name, is_labeled in _SEGMENTS.items():
        segment = dataset.create_segment(segment_name)
        image_paths = glob(os.path.join(root_path, segment_name, "*.jpg"))
        for image_path in image_paths:
            data = Data(image_path)
            if is_labeled:
                data.label.classification = Classification(os.path.basename(image_path)[:3])
            segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""OpenDataset dataloader collections."""

from tensorbay.opendataset.AADB import AADB
from tensorbay.opendataset.AnimalPose import AnimalPose5, AnimalPose7
from tensorbay.opendataset.AnimalsWithAttributes2 import AnimalsWithAttributes2
from tensorbay.opendataset.BDD100K import BDD100K, BDD100K_10K
from tensorbay.opendataset.BDD100K_MOT2020 import BDD100K_MOT2020, BDD100K_MOTS2020
from tensorbay.opendataset.BioIDFace import BioIDFace
from tensorbay.opendataset.BSTLD import BSTLD
from tensorbay.opendataset.CACD import CACD
from tensorbay.opendataset.CADC import CADC
from tensorbay.opendataset.CarConnection import CarConnection
from tensorbay.opendataset.CCPD import CCPD, CCPDGreen
from tensorbay.opendataset.CIHP import CIHP
from tensorbay.opendataset.COCO2017 import COCO2017
from tensorbay.opendataset.CoinImage import CoinImage
from tensorbay.opendataset.CompCars import CompCars
from tensorbay.opendataset.COVID_CT import COVID_CT
from tensorbay.opendataset.COVIDChestXRay import COVIDChestXRay
from tensorbay.opendataset.DAVIS2017 import DAVIS2017SemiSupervised, DAVIS2017Unsupervised
from tensorbay.opendataset.DeepRoute import DeepRoute
from tensorbay.opendataset.DogsVsCats import DogsVsCats
from tensorbay.opendataset.DownsampledImagenet import DownsampledImagenet
from tensorbay.opendataset.Elpv import Elpv
from tensorbay.opendataset.FLIC import FLIC
from tensorbay.opendataset.Flower import Flower17, Flower102
from tensorbay.opendataset.FSDD import FSDD
from tensorbay.opendataset.HalpeFullBody import HalpeFullBody
from tensorbay.opendataset.HardHatWorkers import HardHatWorkers
from tensorbay.opendataset.HeadPoseImage import HeadPoseImage
from tensorbay.opendataset.ImageEmotion import ImageEmotionAbstract, ImageEmotionArtphoto
from tensorbay.opendataset.JHU_CROWD import JHU_CROWD
from tensorbay.opendataset.KenyanFood import KenyanFoodOrNonfood, KenyanFoodType
from tensorbay.opendataset.KylbergTexture import KylbergTexture
from tensorbay.opendataset.LeedsSportsPose import LeedsSportsPose
from tensorbay.opendataset.LIP import LIP
from tensorbay.opendataset.LISATrafficLight import LISATrafficLight
from tensorbay.opendataset.LISATrafficSign import LISATrafficSign
from tensorbay.opendataset.NeolixOD import NeolixOD
from tensorbay.opendataset.Newsgroups20 import Newsgroups20
from tensorbay.opendataset.NightOwls import NightOwls
from tensorbay.opendataset.nuImages import nuImages
from tensorbay.opendataset.nuScenes import nuScenes
from tensorbay.opendataset.OxfordIIITPet import OxfordIIITPet
from tensorbay.opendataset.PASCALContext import PASCALContext
from tensorbay.opendataset.RP2K import RP2K
from tensorbay.opendataset.SCUT_FBP5500 import SCUT_FBP5500
from tensorbay.opendataset.SegTrack import SegTrack
from tensorbay.opendataset.SegTrack2 import SegTrack2
from tensorbay.opendataset.SVHN import SVHN
from tensorbay.opendataset.THCHS30 import THCHS30
from tensorbay.opendataset.THUCNews import THUCNews
from tensorbay.opendataset.TLR import TLR
from tensorbay.opendataset.UAVDT import UAVDT
from tensorbay.opendataset.VOC2012ActionClassification import VOC2012ActionClassification
from tensorbay.opendataset.VOC2012Detection import VOC2012Detection
from tensorbay.opendataset.VOC2012Segmentation import VOC2012Segmentation
from tensorbay.opendataset.WIDER_FACE import WIDER_FACE

__all__ = [
    "nuScenes",
    "AADB",
    "AnimalPose5",
    "AnimalPose7",
    "AnimalsWithAttributes2",
    "BioIDFace",
    "BDD100K",
    "BDD100K_MOT2020",
    "BDD100K_MOTS2020",
    "BDD100K_10K",
    "BSTLD",
    "CACD",
    "CADC",
    "CarConnection",
    "CCPD",
    "CCPDGreen",
    "CIHP",
    "COCO2017",
    "CoinImage",
    "COVIDChestXRay",
    "CompCars",
    "DAVIS2017SemiSupervised",
    "DAVIS2017Unsupervised",
    "DeepRoute",
    "DogsVsCats",
    "DownsampledImagenet",
    "Elpv",
    "FLIC",
    "FSDD",
    "HalpeFullBody",
    "Flower102",
    "Flower17",
    "HardHatWorkers",
    "HeadPoseImage",
    "ImageEmotionAbstract",
    "ImageEmotionArtphoto",
    "JHU_CROWD",
    "KenyanFoodOrNonfood",
    "KenyanFoodType",
    "KylbergTexture",
    "LIP",
    "LISATrafficLight",
    "LISATrafficSign",
    "LeedsSportsPose",
    "NeolixOD",
    "Newsgroups20",
    "NightOwls",
    "OxfordIIITPet",
    "PASCALContext",
    "RP2K",
    "SCUT_FBP5500",
    "SegTrack",
    "SegTrack2",
    "SVHN",
    "nuImages",
    "THCHS30",
    "THUCNews",
    "TLR",
    "UAVDT",
    "WIDER_FACE",
    "COVID_CT",
    "VOC2012Detection",
    "VOC2012ActionClassification",
    "VOC2012Segmentation",
]



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of TLR dataset."""

import os
from collections import defaultdict
from typing import Dict, List
from xml.dom.minidom import parse

from tensorbay.dataset import Data, Dataset
from tensorbay.label import LabeledBox2D
from tensorbay.opendataset._utility import glob

DATASET_NAME = "TLR"


def TLR(path: str) -> Dataset:
    """`TLR <http://www.lara.prd.fr/benchmarks/trafficlightsrecognition>`_ dataset.

    The file structure should like::

        <path>
            root_path/
                Lara3D_URbanSeq1_JPG/
                    frame_011149.jpg
                    frame_011150.jpg
                    frame_<frame_index>.jpg
                    ...
                Lara_UrbanSeq1_GroundTruth_cvml.xml

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    file_paths = glob(os.path.join(root_path, "Lara3D_UrbanSeq1_JPG", "*.jpg"))
    labels = _parse_xml(os.path.join(root_path, "Lara_UrbanSeq1_GroundTruth_cvml.xml"))
    for file_path in file_paths:
        # the image file name looks like:
        # frame_000001.jpg
        frame_index = int(os.path.basename(file_path)[6:-4])
        data = Data(file_path)
        data.label.box2d = labels[frame_index]
        segment.append(data)
    return dataset


def _parse_xml(xml_path: str) -> Dict[int, List[LabeledBox2D]]:
    dom = parse(xml_path)
    frames = dom.documentElement.getElementsByTagName("frame")

    labels = defaultdict(list)

    for frame in frames:
        index = int(frame.getAttribute("number"))
        objects = frame.getElementsByTagName("objectlist")[0]
        for obj in objects.getElementsByTagName("object"):
            box = obj.getElementsByTagName("box")[0]
            box_h = int(box.getAttribute("h"))
            box_w = int(box.getAttribute("w"))
            box_xc = int(box.getAttribute("xc"))
            box_yc = int(box.getAttribute("yc"))

            labels[index].append(
                LabeledBox2D.from_xywh(
                    x=box_xc - box_w // 2,
                    y=box_yc - box_h // 2,
                    width=box_w,
                    height=box_h,
                    category=obj.getElementsByTagName("subtype")[0].firstChild.data,
                )
            )
    return labels



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of COCO2017 dataset."""

import json
import os
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Tuple

import numpy as np

from tensorbay.dataset import Data, Dataset
from tensorbay.label import (
    Label,
    LabeledBox2D,
    LabeledKeypoints2D,
    LabeledMultiPolygon,
    LabeledRLE,
    PanopticMask,
)
from tensorbay.opendataset._utility.glob import glob
from tensorbay.utility.itertools import chunked

try:
    from PIL import Image
except ModuleNotFoundError:
    from tensorbay.opendataset._utility.mocker import Image  # pylint:disable=ungrouped-imports

DATASET_NAME = "COCO2017"
_LABELED_SEGMENT_NAMES = ("train", "val")
_UNLABELED_SEGMENT_NAMES = ("test", "unlabeled")


def COCO2017(path: str) -> Dataset:
    """`COCO2017 <https://cocodataset.org/#home>`_ dataset.

    The file structure should be like::

        <path>
            annotations/
                panoptic_train2017/
                    000000116037.png
                    000000116040.png
                    ...
                panoptic_val2017/
                instances_train2017.json
                instances_val2017.json
                panoptic_train2017.json
                panoptic_val2017.json
                person_keypoints_train2017.json
                person_keypoints_val2017.json
            train2017/
                000000116026.jpg
                000000116031.jpg
                ...
            test2017/
            val2017/
            unlabeled2017/

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name in _LABELED_SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        annotation_path = os.path.join(root_path, "annotations")
        task_information = _get_information(annotation_path, segment_name)
        categories = task_information["categories"]
        for image_path in glob(os.path.join(root_path, f"{segment_name}2017", "*.jpg")):
            data = Data(image_path)
            image_stem = os.path.splitext(os.path.basename(image_path))[0]
            image_id = int(image_stem)

            label = _get_instance_label(
                task_information["instances_annotations"], image_id, categories
            )
            label.keypoints2d = _get_keypoints2d(
                task_information["person_keypoints_annotations"], image_id, categories
            )
            label.box2d.extend(
                _get_panoptic_box2d(task_information["panoptic_annotations"], image_id, categories)
            )
            label.panoptic_mask = _get_panoptic_mask(
                annotation_path,
                segment_name,
                image_stem,
                task_information["panoptic_annotations"],
                image_id,
            )
            data.label = label
            segment.append(data)

    for segment_name in _UNLABELED_SEGMENT_NAMES:
        segment = dataset.create_segment(segment_name)
        for image_path in glob(os.path.join(root_path, f"{segment_name}2017", "*.jpg")):
            segment.append(Data(image_path))

    return dataset


def _get_information(annotation_path: str, segment_name: str) -> Dict[str, Any]:
    task_information: Dict[str, Any] = {}
    for task in ("instances", "person_keypoints", "panoptic"):
        with open(
            os.path.join(annotation_path, f"{task}_{segment_name}2017.json"), "r", encoding="utf-8"
        ) as fp:
            file_json = json.load(fp)

        task_annotation: DefaultDict[int, Any] = defaultdict(list)
        for annotation in file_json["annotations"]:
            task_annotation[annotation["image_id"]].append(annotation)
        task_information[f"{task}_annotations"] = task_annotation

        if task == "panoptic":
            task_information["categories"] = {
                category["id"]: f"{category['supercategory']}.{category['name']}"
                for category in file_json["categories"]
            }
    return task_information


def _get_instance_label(
    instances_annotations: Dict[int, Any], image_id: int, categories: Dict[int, str]
) -> Label:
    label: Label = Label()
    label.box2d = []
    label.multi_polygon = []
    label.rle = []
    if image_id not in instances_annotations:
        return label

    for annotation in instances_annotations[image_id]:
        category = categories[annotation["category_id"]]
        label.box2d.append(LabeledBox2D.from_xywh(*annotation["bbox"], category=category))
        if annotation["iscrowd"] == 0:
            points = [chunked(coordinates, 2) for coordinates in annotation["segmentation"]]
            label.multi_polygon.append(LabeledMultiPolygon(points, category=category))
        else:
            label.rle.append(LabeledRLE(annotation["segmentation"]["counts"], category=category))
    return label


def _get_keypoints2d(
    person_keypoints_annotations: Dict[int, Any], image_id: int, categories: Dict[int, str]
) -> List[LabeledKeypoints2D]:
    if image_id not in person_keypoints_annotations:
        return []

    keypoints2d: List[LabeledKeypoints2D] = []
    for annotation in person_keypoints_annotations[image_id]:
        points = chunked(annotation["keypoints"], 3)
        category = categories[annotation["category_id"]]
        keypoints2d.append(LabeledKeypoints2D(points, category=category))

    return keypoints2d


def _get_panoptic_box2d(
    panoptic_annotations: Dict[int, Any], image_id: int, categories: Dict[int, str]
) -> List[LabeledBox2D]:
    if image_id not in panoptic_annotations:
        return []

    box2d: List[LabeledBox2D] = []
    for annotation in panoptic_annotations[image_id]:
        for segment_info in annotation["segments_info"]:
            # category_id 1-91 are thing categories from the detection task
            # category_id 92-200 are stuff categories from the stuff task
            if segment_info["category_id"] > 91:
                category = categories[segment_info["category_id"]]
                box2d.append(LabeledBox2D.from_xywh(*segment_info["bbox"], category=category))
    return box2d


def _get_panoptic_mask(
    annotation_path: str,
    segment_name: str,
    image_stem: str,
    panoptic_annotations: Dict[int, Any],
    image_id: int,
) -> PanopticMask:
    compress_pixel, new_mask_path = _save_mask(annotation_path, segment_name, image_stem)

    category_ids = {}
    annotation = panoptic_annotations[image_id][0]
    for info in annotation["segments_info"]:
        category_ids[compress_pixel[info["id"]]] = info["category_id"]
    panoptic_mask = PanopticMask(local_path=new_mask_path)
    panoptic_mask.all_category_ids = category_ids
    return panoptic_mask


def _save_mask(
    annotation_path: str, segment_name: str, image_stem: str
) -> Tuple[Dict[int, int], str]:
    new_file_path = os.path.join(annotation_path, f"new_panoptic_{segment_name}2017")
    os.makedirs(new_file_path, exist_ok=True)
    mask_name = f"{image_stem}.png"

    array = np.array(
        Image.open(os.path.join(annotation_path, f"panoptic_{segment_name}2017", mask_name)),
        dtype=int,
    )
    mask_array = array[:, :, 0] + array[:, :, 1] * 256 + array[:, :, 2] * 256 * 256

    array_elements = np.unique(mask_array)
    array_elements = np.delete(array_elements, np.where(array_elements == 0))
    map_of_pixel = dict(zip(array_elements, range(1, len(array_elements) + 1)))
    map_of_pixel[0] = 0
    mask_array = np.vectorize(map_of_pixel.get)(mask_array)

    new_mask_path = os.path.join(new_file_path, mask_name)
    Image.fromarray(np.uint8(mask_array)).save(new_mask_path)
    return map_of_pixel, new_mask_path



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of the ImageEmotionAbstract dataset and the ImageEmotionArtphoto dataset."""

from tensorbay.opendataset.ImageEmotion.loader import ImageEmotionAbstract, ImageEmotionArtphoto

__all__ = ["ImageEmotionAbstract", "ImageEmotionArtphoto"]



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of SegTrack2 dataset."""

import os
from typing import Callable, Dict

import numpy as np

from tensorbay.dataset import Data, Dataset
from tensorbay.label import InstanceMask
from tensorbay.opendataset._utility import glob

try:
    from PIL import Image
except ModuleNotFoundError:
    from tensorbay.opendataset._utility.mocker import Image  # pylint:disable=ungrouped-imports

DATASET_NAME = "SegTrack2"

_SEGMENTS_INFO = {
    1: ("bird_of_paradise", "birdfall", "frog", "girl", "monkey", "parachute", "soldier", "worm"),
    2: ("bmx", "cheetah", "drift", "hummingbird", "monkeydog"),
    6: ("penguin",),
}
_FILENAME_REFORMATTERS = (
    lambda filename: filename,
    lambda filename: f"{os.path.splitext(filename)[0]}.png",
)
_MASK_GETTER = Callable[[str, str, str, int, Callable[[str], str]], InstanceMask]


def SegTrack2(path: str) -> Dataset:
    """`SegTrack2 <https://web.engr.oregonstate.edu/~lif/SegTrack2/dataset.html>`_ dataset.

    The file structure of SegTrack looks like::

        <path>
            GroundTruth/
                bird_of_paradise/
                    bird_of_paradise_00000.png
                    ...
                bmx/
                    1/
                        bmx_06668.png
                        ...
                    2/
                        bmx_06668.png
                        ...
                ...
            JPEGImages/
                bird_of_paradise/
                    bird_of_paradise_00000.png
                    ...
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.join(os.path.abspath(os.path.expanduser(path)))
    image_dir = os.path.join(root_path, "JPEGImages")
    original_mask_dir = os.path.join(root_path, "GroundTruth")
    mask_dir = os.path.join(root_path, "Masks")
    dataset = Dataset(DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    for instance_num, segment_names in _SEGMENTS_INFO.items():
        for segment_name in segment_names:
            segment = dataset.create_segment(segment_name)
            mask_subdir = os.path.join(mask_dir, segment_name)
            os.makedirs(mask_subdir, exist_ok=True)
            all_mask_subdirs = {
                "original_mask_subdir": os.path.join(original_mask_dir, segment_name),
                "mask_subdir": mask_subdir,
            }

            filename_reformatter = (
                _FILENAME_REFORMATTERS[1]
                if segment_name in {"penguin", "monkeydog"}
                else _FILENAME_REFORMATTERS[0]
            )
            mask_getter: _MASK_GETTER = (
                _get_cheetah_instance_mask if segment_name == "cheetah" else _get_instance_mask
            )
            for image_path in glob(os.path.join(image_dir, segment_name, "*")):
                segment.append(
                    _get_data(
                        image_path,
                        all_mask_subdirs,
                        instance_num,
                        filename_reformatter,
                        mask_getter,
                    )
                )
    return dataset


def _get_data(
    image_path: str,
    all_mask_subdirs: Dict[str, str],
    instance_num: int,
    filename_reformatter: Callable[[str], str],
    mask_getter: _MASK_GETTER,
) -> Data:
    data = Data(image_path)
    data.label.instance_mask = mask_getter(
        image_path,
        all_mask_subdirs["mask_subdir"],
        all_mask_subdirs["original_mask_subdir"],
        instance_num,
        filename_reformatter,
    )
    return data


def _get_instance_mask(
    image_path: str,
    mask_subdir: str,
    original_mask_subdir: str,
    instance_num: int,
    filename_reformatter: Callable[[str], str],
) -> InstanceMask:
    filename = filename_reformatter(os.path.basename(image_path))
    mask_path = os.path.join(mask_subdir, f"{os.path.splitext(filename)[0]}.png")
    if instance_num == 1:
        mask = _get_reformatted_mask(os.path.join(original_mask_subdir, filename))
    else:
        mask = _get_reformatted_mask(os.path.join(original_mask_subdir, "1", filename))
        for instance_id in range(2, instance_num + 1):
            alter_mask = np.array(
                Image.open(os.path.join(original_mask_subdir, str(instance_id), filename)),
            )[:, :, 0]
            mask[alter_mask == 255] = instance_id

    Image.fromarray(mask).save(mask_path)
    return InstanceMask(mask_path)


def _get_cheetah_instance_mask(
    image_path: str, mask_subdir: str, original_mask_subdir: str, _: int, __: Callable[[str], str]
) -> InstanceMask:
    filename = os.path.basename(image_path)
    new_filename = f"{os.path.splitext(filename)[0]}.png"
    mask_path = os.path.join(mask_subdir, new_filename)
    mask = _get_reformatted_mask(os.path.join(original_mask_subdir, "1", filename))
    alter_mask = np.array(
        Image.open(os.path.join(original_mask_subdir, "2", new_filename)),
    )[:, :, 0]
    mask[alter_mask == 255] = 2

    Image.fromarray(mask).save(mask_path)
    return InstanceMask(mask_path)


def _get_reformatted_mask(original_mask_path: str) -> np.ndarray:
    mask = np.array(Image.open(original_mask_path))[:, :, 0]
    # reformat mask
    # from {background: 0, overlap: 1~254, target: 255}
    # to {background: 0, target: 1, overlap: 255}
    overlap = np.logical_and(mask > 0, mask < 255)
    mask[mask == 255] = 1
    mask[overlap] = 255
    return mask



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Statistical.

:class:`Profile` is a class used to save statistical summary.

"""

import csv
import json
import math
import time
from collections import OrderedDict, defaultdict
from functools import wraps
from itertools import chain
from multiprocessing import Manager
from multiprocessing.managers import SyncManager
from types import MethodType
from typing import Any, Callable, DefaultDict, Dict, Optional, TypeVar
from urllib.parse import urlparse

from requests.models import Response
from requests_toolbelt.multipart.encoder import FileWrapper, MultipartEncoder

from tensorbay.client.requests import Client
from tensorbay.utility import RemoteFileMixin, UserResponse

_Callable = TypeVar("_Callable", bound=Callable[..., Response])
_OpenCallable = TypeVar("_OpenCallable", bound=Callable[..., UserResponse])
_ReadCallable = Callable[..., bytes]

_COLUMNS = OrderedDict(
    totalTime="{:<20.3f}",
    callNumber="{:<20}",
    avgTime="{:<20.3f}",
    totalResponseLength="{:<20}",
    totalFileSize="{:<13}",
    speed="{:<13}",
)
_TITLE = f"{'totalTime (s)':<20}{' |callNumber':<18}\
    {' |avgTime (s)':<22}{' |totalResponseLength':<22}{' |totalFileSize':<15}{' |speed':<15}"
_PATH_PREFIX = "/gatewayv2/tensorbay-open-api/v1/"
_UNITS = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")


def format_size(size: float) -> str:
    """Format a byte count as a human readable file size.

    Arguments:
        size: The size to format in bytes.

    Returns:
        The human readable string.

    """
    if size == 0:
        return "0B"

    index = int(math.log(size, 1024))
    number = pow(1024, index)
    return f"{size/number:.2f}{_UNITS[index]}"


class Profile:
    """This is a class used to save statistical summary."""

    _manager: SyncManager
    _summary: Dict[str, DefaultDict[str, Any]]

    def __init__(self) -> None:
        self.do_function = Client.do
        self.urlopen = RemoteFileMixin._urlopen

    def __enter__(self) -> "Profile":
        self.start()
        return self

    def __exit__(self, *_: Any) -> None:
        self.stop()

    def _calculate_and_format(self) -> None:
        for key, value in self._summary.items():
            value["avgTime"] = value["totalTime"] / value["callNumber"]
            value["totalFileSize"] = format_size(value["rawTotalFileSize"])
            value["speed"] = f'{format_size(value["rawTotalFileSize"] / value["totalTime"])}/s'
            self._summary[key] = value

    @staticmethod
    def _format_string(path: str = "Path", item: Optional[Dict[str, float]] = None) -> str:
        content = [f"|{path:<63}"]
        if item is None:
            content.append(_TITLE)
        else:
            content.extend(value.format(item[key]) for key, value in _COLUMNS.items())
        content.append("\n")
        return " |".join(content)

    @staticmethod
    def _get_file_size(data: MultipartEncoder) -> int:
        for part in data.parts:
            body = part.body
            if isinstance(body, FileWrapper):
                return body.len  # type: ignore[no-any-return]

        return 0

    def _save_to_txt(self, path: str) -> None:
        with open(path, "w", encoding="utf-8") as fp:
            fp.write(self._format_string())
            for url, item in self._summary.items():
                fp.write(self._format_string(url, item))

    def _save_to_csv(self, path: str) -> None:
        with open(path, "w", encoding="utf-8") as fp:
            writer = csv.writer(fp)
            writer.writerow(chain(["Path"], _COLUMNS.keys()))
            for post, item in self._summary.items():
                writer.writerow(chain([post], (item[key] for key in _COLUMNS)))

    def _save_to_json(self, path: str) -> None:
        summary = self._summary if isinstance(self._summary, dict) else self._summary.copy()
        with open(path, "w", encoding="utf-8") as fp:
            json.dump(summary, fp, indent=4)

    def _statistical(self, func: _Callable) -> _Callable:
        """A decorator to record the running time, response length ..etc of the function.

        Arguments:
            func: The Client.do() function.

        Returns:
            The wrapped function.

        """

        @wraps(func)
        def wrapper(client: Client, method: str, url: str, **kwargs: Any) -> Response:
            path = urlparse(url).path
            part = path.replace(_PATH_PREFIX, "") if path else url
            key = f"[{method}] {part}"
            data = kwargs.get("data")
            file_size = self._get_file_size(data) if isinstance(data, MultipartEncoder) else 0
            response = func(client, method, url, **kwargs)
            self._update(key, len(response.content), response.elapsed.total_seconds(), file_size)

            return response

        return wrapper  # type: ignore[return-value]

    def _statistical_read(self, download_path: str) -> _ReadCallable:
        def wrapper(response: UserResponse, amt: Optional[int] = None) -> bytes:
            start_time = time.time()
            content = UserResponse.read(response, amt)
            self._update(download_path, 0, time.time() - start_time, len(content))
            return content

        return wrapper

    def _statistical_open(self, func: _OpenCallable) -> _OpenCallable:
        @wraps(func)
        def wrapper(obj: RemoteFileMixin) -> UserResponse:
            netloc = urlparse(obj.url.get()).netloc  # type: ignore[union-attr]
            download_path = f"[GET] {netloc}/*"

            start_time = time.time()
            fp = func(obj)
            cost_time = time.time() - start_time

            item = self._summary.get(download_path, defaultdict(int))
            item["totalTime"] += cost_time
            self._summary[download_path] = item
            setattr(fp, "read", MethodType(self._statistical_read(download_path), fp))
            return fp

        return wrapper  # type: ignore[return-value]

    def _update(self, key: str, response_length: int, cost_time: float, file_size: int) -> None:
        """Update the giving information into summary.

        Arguments:
            key: the post method and url.
            response_length: response length of the post.
            cost_time: cost time of the post.
            file_size: the upload file size.

        """
        item = self._summary.get(key, defaultdict(int))
        item["totalTime"] += cost_time
        item["callNumber"] += 1
        item["totalResponseLength"] += response_length
        item["rawTotalFileSize"] += file_size
        self._summary[key] = item

    def save(self, path: str, file_type: str = "txt") -> None:
        """Save the statistical summary into a file.

        Arguments:
            path: The file local path.
            file_type: Type of the save file, only support 'txt', 'json', 'csv'.

        """
        self._calculate_and_format()
        writers = {"txt": self._save_to_txt, "json": self._save_to_json, "csv": self._save_to_csv}
        writers[file_type](path)

    def start(self, multiprocess: bool = False) -> None:
        """Start statistical record.

        Arguments:
            multiprocess: Whether the records is in a multi-process environment.

        """
        if multiprocess:
            self._manager = Manager()
            self._summary = self._manager.dict()
        else:
            self._summary = {}
        setattr(Client, "do", self._statistical(self.do_function))
        setattr(RemoteFileMixin, "_urlopen", self._statistical_open(self.urlopen))

    def stop(self) -> None:
        """Stop statistical record."""
        setattr(Client, "do", self.do_function)
        setattr(RemoteFileMixin, "_urlopen", self.urlopen)
        if hasattr(self, "_manager"):
            self._summary = self._summary.copy()
            self._manager.shutdown()
            delattr(self, "_manager")


profile = Profile()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class Client and method multithread_upload.

:class:`Client` can send POST, PUT, and GET requests to the TensorBay Dataset Open API.

:meth:`multithread_upload` creates a multi-thread framework for uploading.

"""

import logging
from concurrent.futures import FIRST_EXCEPTION, ThreadPoolExecutor, wait
from queue import Queue
from threading import Lock
from typing import Any, Callable, Generic, Iterable, Optional, Tuple, TypeVar
from urllib.parse import urljoin
from uuid import uuid4

from requests.models import Response

from tensorbay.__version__ import __version__
from tensorbay.exception import ResponseError, ResponseErrorDistributor
from tensorbay.utility import Tqdm, config, get_session

logger = logging.getLogger(__name__)


class Client:
    """This class defines :class:`Client`.

    :class:`Client` defines the client that saves the user and URL information
    and supplies basic call methods that will be used by derived clients,
    such as sending GET, PUT and POST requests to TensorBay Open API.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti gas website.

    """

    _DEFAULT_URL_CN = "https://gas.graviti.cn/"
    _DEFAULT_URL_COM = "https://gas.graviti.com/"

    def __init__(self, access_key: str, url: str = "") -> None:
        if access_key.startswith("Accesskey-"):
            url = url if url else Client._DEFAULT_URL_CN
        elif access_key.startswith("ACCESSKEY-"):
            url = url if url else Client._DEFAULT_URL_COM
        else:
            raise TypeError("Wrong accesskey format!")

        if not url.startswith("https://"):
            raise TypeError("Invalid url, only support url starts with 'https://'")

        self.gateway_url = urljoin(url, "gatewayv2/")
        self.access_key = access_key

        self._open_api = urljoin(self.gateway_url, "tensorbay-open-api/v1/")

    def _url_make(self, section: str, dataset_id: str = "") -> str:
        """Generate Open API URL.

        Arguments:
            section: The section of the request.
            dataset_id: Dataset ID.

        Returns:
             Open API URL.

        """
        if dataset_id:
            dataset_url = urljoin(self._open_api, "datasets/")
            if section:
                url = urljoin(urljoin(dataset_url, dataset_id + "/"), section)
            else:
                url = urljoin(dataset_url, dataset_id)
        else:
            if section:
                url = urljoin(self._open_api, section)
            else:
                url = urljoin(self._open_api, "datasets")
        return url

    @staticmethod
    def do(method: str, url: str, **kwargs: Any) -> Response:  # pylint: disable=invalid-name
        """Send a request.

        Arguments:
            method: The method of the request.
            url: The URL of the request.
            **kwargs: Extra keyword arguments to send in the GET request.

        Returns:
            Response of the request.

        """
        return get_session().request(method=method, url=url, **kwargs)

    def open_api_do(
        self, method: str, section: str, dataset_id: str = "", **kwargs: Any
    ) -> Response:
        """Send a request to the TensorBay Open API.

        Arguments:
            method: The method of the request.
            section: The section of the request.
            dataset_id: Dataset ID.
            **kwargs: Extra keyword arguments to send in the POST request.

        Raises:
            ResponseError: When the status code OpenAPI returns is unexpected.

        Returns:
            Response of the request.

        """
        headers = kwargs.setdefault("headers", {})
        headers["X-Token"] = self.access_key
        headers[
            "X-Source"
        ] = f"{config._x_source}/{__version__}"  # pylint: disable=protected-access
        headers["X-Request-Id"] = uuid4().hex

        try:
            return self.do(method=method, url=self._url_make(section, dataset_id), **kwargs)
        except ResponseError as error:
            response = error.response
            error_code = response.json()["code"]
            raise ResponseErrorDistributor.get(error_code, ResponseError)(
                response=response
            ) from None


_T = TypeVar("_T")
_R = TypeVar("_R")


def multithread_upload(
    function: Callable[[_T], Optional[_R]],
    arguments: Iterable[_T],
    *,
    callback: Optional[Callable[[Tuple[_R, ...]], None]] = None,
    jobs: int = 1,
    pbar: Tqdm,
) -> None:
    """Multi-thread upload framework.

    Arguments:
        function: The upload function.
        arguments: The arguments of the upload function.
        callback: The callback function.
        jobs: The number of the max workers in multi-thread uploading procession.
        pbar: The :class:`Tqdm` instance for showing the upload process bar.

    """
    with ThreadPoolExecutor(jobs) as executor:
        if callback is not None:
            multi_callback = MultiCallbackTask(function=function, callback=callback)
            function = multi_callback.work
        futures = [executor.submit(function, argument) for argument in arguments]

        for future in futures:
            future.add_done_callback(pbar.update_callback)

        done, not_done = wait(futures, return_when=FIRST_EXCEPTION)

        if callback is not None:
            multi_callback.last_callback()

        for future in not_done:
            future.cancel()
        for future in done:
            future.result()


class MultiCallbackTask(Generic[_T, _R]):
    """A class for callbacking in multi-thread work.

    Arguments:
        function: The function of a single thread.
        callback: The callback function.
        size: The size of the task queue to send a callback.

    """

    def __init__(
        self,
        *,
        function: Callable[[_T], Optional[_R]],
        callback: Callable[[Tuple[_R, ...]], None],
        size: int = 50,
    ) -> None:
        self._lock = Lock()
        self._function = function
        self._callback = callback
        self._size = size
        self._arguments: Queue[_R] = Queue()  # pylint: disable=unsubscriptable-object

    def work(self, argument: _T) -> None:
        """Do the work of a single thread.

        Arguments:
            argument: The argument of the function.

        """
        callback_info = self._function(argument)
        if callback_info is not None:
            self._arguments.put(callback_info)
        with self._lock:
            if self._arguments.qsize() >= self._size:
                callback_arguments = tuple(self._arguments.get() for _ in range(self._size))
            else:
                callback_arguments = ()

        if callback_arguments:
            self._callback(callback_arguments)

    def last_callback(self) -> None:
        """Send the last callback when all works have been done."""
        callback_arguments = tuple(self._arguments.get() for _ in range(self._arguments.qsize()))
        if callback_arguments:
            self._callback(callback_arguments)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of THUCNews dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME = "THUCNews"


def THUCNews(path: str) -> Dataset:
    """`THUCNews <http://thuctc.thunlp.org/>`_ dataset.

    The folder structure should be like::

        <path>
            <category>/
                0.txt
                1.txt
                2.txt
                3.txt
                ...
            <category>/
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    for category in dataset.catalog.classification.categories.keys():
        text_paths = glob(os.path.join(root_path, category, "*.txt"))
        for text_path in text_paths:
            data = Data(text_path)
            data.label.classification = Classification(category)

            segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of ImageEmotionAbstract dataset and ImageEmotionArtPhoto dataset."""

import csv
import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME_ABSTRACT = "ImageEmotionAbstract"
DATASET_NAME_ARTPHOTO = "ImageEmotionArtPhoto"


def ImageEmotionAbstract(path: str) -> Dataset:
    """`Image Emotion-abstract <https://www.imageemotion.org/>`_ dataset.

    The file structure should be like::

        <path>
            ABSTRACT_groundTruth.csv
            abstract_xxxx.jpg
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME_ABSTRACT)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_abstract.json"))
    segment = dataset.create_segment()

    csv_path = os.path.join(root_path, "ABSTRACT_groundTruth.csv")
    with open(csv_path, "r", encoding="utf-8") as fp:
        reader = csv.DictReader(fp)
        reader.fieldnames = [
            field.strip("'") for field in reader.fieldnames  # type:ignore[union-attr]
        ]

        for row in reader:
            image_path = os.path.join(root_path, row.pop("").strip("'"))

            data = Data(image_path)
            values = {key: int(value) for key, value in row.items()}

            data.label.classification = Classification(attributes=values)
            segment.append(data)

    return dataset


def ImageEmotionArtphoto(path: str) -> Dataset:
    """`Image Emotion-art Photo <https://www.imageemotion.org/>`_ dataset.

    The file structure should be like::

        <path>
            <filename>.jpg
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME_ARTPHOTO)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_artphoto.json"))
    segment = dataset.create_segment()

    image_paths = glob(os.path.join(root_path, "*.jpg"))

    for image_path in image_paths:
        image_category = os.path.basename(image_path).split("_", 1)[0]

        data = Data(image_path)
        data.label.classification = Classification(category=image_category)
        segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Label.

A :class:`~.tensorbay.dataset.data.Data` instance contains one or several types of labels,
all of which are stored in :attr:`~tensorbay.dataset.data.Data.label`.

Different label types correspond to different label classes classes.

.. table:: label classes
   :widths: auto

   ================================ ===================================
   label classes                    explaination
   ================================ ===================================
   :class:`.Classification`         classification type of label
   :class:`.LabeledBox2D`           2D bounding box type of label
   :class:`.LabeledBox3D`           3D bounding box type of label
   :class:`.LabeledPolygon`         polygon type of label
   :class:`.LabeledMultiPolygon`    polygon lists type of label
   :class:`.LabeledRLE`             rle mask type of label
   :class:`.LabeledPolyline2D`      2D polyline type of label
   :class:`.LabeledMultiPolyline2D` 2D polyline lists type of label
   :class:`.LabeledKeypoints2D`     2D keypoints type of label
   :class:`.LabeledSentence`        transcripted sentence type of label
   ================================ ===================================

"""

from functools import partial
from typing import Any, Dict, Iterator, List, Type, TypeVar

from tensorbay.label.label_box import LabeledBox2D, LabeledBox3D
from tensorbay.label.label_classification import Classification
from tensorbay.label.label_keypoints import LabeledKeypoints2D
from tensorbay.label.label_mask import (
    InstanceMaskBase,
    PanopticMaskBase,
    RemoteInstanceMask,
    RemotePanopticMask,
    RemoteSemanticMask,
    SemanticMaskBase,
)
from tensorbay.label.label_polygon import LabeledMultiPolygon, LabeledPolygon, LabeledRLE
from tensorbay.label.label_polyline import LabeledMultiPolyline2D, LabeledPolyline2D
from tensorbay.label.label_sentence import LabeledSentence
from tensorbay.utility import AttrsMixin, ReprMixin, ReprType, attr, common_loads, upper

_ERROR_MESSAGE = "The '{attr_name}' label is not provided in this data"
_attr = partial(attr, is_dynamic=True, key=upper, error_message=_ERROR_MESSAGE)
_mask_attr = partial(_attr, dumper=lambda self: self.get_callback_body())


class Label(ReprMixin, AttrsMixin):  # pylint: disable=too-many-instance-attributes
    """This class defines :attr:`~tensorbay.dataset.data.Data.label`.

    It contains growing types of labels referring to different tasks.

    Examples:
        >>> from tensorbay.label import Classification
        >>> label = Label()
        >>> label.classification = Classification("example_category", {"example_attribute1": "a"})
        >>> label
        Label(
          (classification): Classification(
            (category): 'example_category',
            (attributes): {...}
          )
        )

    """

    _T = TypeVar("_T", bound="Label")

    _repr_type = ReprType.INSTANCE
    _repr_maxlevel = 2

    classification: Classification = _attr()
    box2d: List[LabeledBox2D] = _attr()
    box3d: List[LabeledBox3D] = _attr()
    polygon: List[LabeledPolygon] = _attr()
    polyline2d: List[LabeledPolyline2D] = _attr()
    multi_polyline2d: List[LabeledMultiPolyline2D] = _attr()
    rle: List[LabeledRLE] = _attr()
    keypoints2d: List[LabeledKeypoints2D] = _attr()
    multi_polygon: List[LabeledMultiPolygon] = _attr()
    sentence: List[LabeledSentence] = _attr()
    semantic_mask: SemanticMaskBase = _mask_attr(loader=RemoteSemanticMask.from_response_body)
    instance_mask: InstanceMaskBase = _mask_attr(loader=RemoteInstanceMask.from_response_body)
    panoptic_mask: PanopticMaskBase = _mask_attr(loader=RemotePanopticMask.from_response_body)

    def __bool__(self) -> bool:
        return any(hasattr(self, key) for key in self._attrs_fields)

    @property
    def _repr_attrs(self) -> Iterator[str]:  # type: ignore[override]
        yield from self._attrs_fields

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads data from a dict containing the labels information.

        Arguments:
            contents: A dict containing the labels information.

        Returns:
            A :class:`Label` instance containing labels information from the given dict.

        Examples:
            >>> contents = {
            ...     "CLASSIFICATION": {
            ...         "category": "example_category",
            ...         "attributes": {"example_attribute1": "a"}
            ...     }
            ... }
            >>> Label.loads(contents)
            Label(
              (classification): Classification(
                (category): 'example_category',
                (attributes): {...}
              )
            )

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all labels into a dict.

        Returns:
            Dumped labels dict.

        Examples:
            >>> from tensorbay.label import Classification
            >>> label = Label()
            >>> label.classification = Classification("category1", {"attribute1": "a"})
            >>> label.dumps()
            {'CLASSIFICATION': {'category': 'category1', 'attributes': {'attribute1': 'a'}}}

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of KenyanFoodType dataset and KenyanFoodOrNonfood dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME_FOOD_TYPE = "KenyanFoodType"
DATASET_NAME_FOOD_OR_NONFOOD = "KenyanFoodOrNonfood"
SEGMENTS_FOOD_TYPE = ["test", "train", "val"]
SEGMENTS_FOOD_OR_NONFOOD = {"test": "test.txt", "train": "train.txt"}


def KenyanFoodOrNonfood(path: str) -> Dataset:
    """`Kenyan Food or Nonfood <https://github.com/monajalal/Kenyan-Food>`_ dataset.

    The file structure should be like::

        <path>
            images/
                food/
                    236171947206673742.jpg
                    ...
                nonfood/
                    168223407.jpg
                    ...
            data.csv
            split.py
            test.txt
            train.txt

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME_FOOD_OR_NONFOOD)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_food_or_nonfood.json"))

    for segment_name, filename in SEGMENTS_FOOD_OR_NONFOOD.items():
        segment = dataset.create_segment(segment_name)
        with open(os.path.join(root_path, filename), encoding="utf-8") as fp:
            for image_path in fp:
                image_path = os.path.join(root_path, image_path)
                data = Data(image_path.strip())
                category = image_path.split("/")[1]
                data.label.classification = Classification(category)
                segment.append(data)
    return dataset


def KenyanFoodType(path: str) -> Dataset:
    """`Kenyan Food Type <https://github.com/monajalal/Kenyan-Food>`_ dataset.

    The file structure should be like::

        <path>
            test.csv
            test/
                bhaji/
                    1611654056376059197.jpg
                    ...
                chapati/
                    1451497832469337023.jpg
                    ...
                ...
            train/
                bhaji/
                    190393222473009410.jpg
                    ...
                chapati/
                    1310641031297661755.jpg
                    ...
            val/
                bhaji/
                    1615408264598518873.jpg
                    ...
                chapati/
                    1553618479852020228.jpg
                    ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME_FOOD_TYPE)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_food_type.json"))

    for segment_name in SEGMENTS_FOOD_TYPE:
        segment = dataset.create_segment(segment_name)
        segment_path = os.path.join(root_path, segment_name)
        for category in sorted(os.listdir(segment_path)):
            image_paths = glob(os.path.join(segment_path, category, "*.jpg"))
            label = Classification(category)
            for image_path in image_paths:
                data = Data(image_path)
                data.label.classification = label
                segment.append(data)
    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Hard Hat Workers dataset."""

from tensorbay.opendataset.HardHatWorkers.loader import HardHatWorkers

__all__ = ["HardHatWorkers"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of CCPD dataset."""

import os
from typing import Dict, Iterator, List, Union

from tensorbay.dataset import Data, Dataset
from tensorbay.label import LabeledPolygon
from tensorbay.opendataset._utility.glob import glob

_PROVINCES = (
    "皖",
    "沪",
    "津",
    "渝",
    "冀",
    "晋",
    "蒙",
    "辽",
    "吉",
    "黑",
    "苏",
    "浙",
    "京",
    "闽",
    "赣",
    "鲁",
    "豫",
    "鄂",
    "湘",
    "粤",
    "桂",
    "琼",
    "川",
    "贵",
    "云",
    "藏",
    "陕",
    "甘",
    "青",
    "宁",
    "新",
    "警",
    "学",
)

_CODES = (
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
)

_ADS = (
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
)

_CCPD_SEGMENTS = {
    "other": ("np", "weather"),
    "train": ("base",),
    "val": ("base",),
    "test": ("blur", "challenge", "db", "fn", "rotate", "tilt"),
}

_CCPDGREEN_SEGMENTS = ("train", "val", "test")

DATASET_NAME_CCPD = "CCPD"
DATASET_NAME_CCPDGREEN = "CCPDGreen"


def CCPD(path: str) -> Dataset:
    """`CCPD <https://github.com/detectRecog/CCPD>`_ dataset.

    The file structure should be like::

        <path>
            ccpd_np/
                1005.jpg
                1019.jpg
                ...
            ccpd_base/
                00205459770115-90_85-352&516_448&547- \
                444&547_368&549_364&517_440&515-0_0_22_10_26_29_24-128-7.jpg
                00221264367816-91_91-283&519_381&553- \
                375&551_280&552_285&514_380&513-0_0_7_26_17_33_29-95-9.jpg
                ...
            ccpd_blur/
            ccpd_challenge/
            ccpd_db/
            ccpd_fn/
            ccpd_rotate/
            ccpd_tilt/
            ccpd_weather/
            LICENSE
            README.md
            splits/
                ccpd_blur.txt
                ccpd_challenge.txt
                ccpd_db.txt
                ccpd_fn.txt
                ccpd_rotate.txt
                ccpd_tilt.txt
                test.txt
                train.txt
                val.txt

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME_CCPD)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_head, segment_tails in _CCPD_SEGMENTS.items():
        for segment_tail in segment_tails:
            segment_name = f"{segment_head}-{segment_tail}"
            segment = dataset.create_segment(segment_name)
            get_polygons = _get_polygons if segment_name != "other-np" else lambda _: []
            for image_path in _get_ccpd_image_path(root_path, segment_head, segment_tail):
                data = Data(image_path)
                data.label.polygon = get_polygons(image_path)
                segment.append(data)
    return dataset


def CCPDGreen(path: str) -> Dataset:
    """`CCPDGreen <https://github.com/detectRecog/CCPD>`_ dataset.

    The file structure should be like::

        <path>
            ccpd_green/
                train/
                test/
                val/

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.join(os.path.abspath(os.path.expanduser(path)), "ccpd_green")

    dataset = Dataset(DATASET_NAME_CCPDGREEN)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name in _CCPDGREEN_SEGMENTS:
        segment = dataset.create_segment(segment_name)
        for image_path in glob(os.path.join(root_path, segment_name, "*.jpg")):
            data = Data(image_path)
            data.label.polygon = _get_polygons(image_path)
            segment.append(data)
    return dataset


def _get_license_plate(license_index: str) -> str:
    indexes = map(int, license_index.split("_"))
    return (
        f"{_PROVINCES[next(indexes)]}{_CODES[next(indexes)]}"
        f"{''.join(_ADS[index] for index in indexes)}"
    )


def _get_polygons(image_path: str) -> List[LabeledPolygon]:
    attributes: Dict[str, Union[int, str]] = {}
    annotations = os.path.splitext(os.path.basename(image_path))[0].split("-", 6)
    points = (map(int, point.split("&")) for point in annotations[3].split("_", 3))
    tilt_degree = annotations[1].split("_", 1)
    attributes["horizontal_tilt"] = int(tilt_degree[0])
    attributes["vertical_tilt"] = int(tilt_degree[1])
    attributes["license_plate"] = _get_license_plate(annotations[4])
    attributes["brightness"] = int(annotations[5])
    attributes["blurriness"] = int(annotations[6])
    return [LabeledPolygon(points, attributes=attributes)]


def _get_ccpd_image_path(root_path: str, segment_head: str, segment_tail: str) -> Iterator[str]:
    if segment_tail == "base":
        file_path = os.path.join(root_path, "splits", f"{segment_head}.txt")
        with open(file_path, "r", encoding="utf-8") as fp:
            for image_path in fp:
                yield os.path.join(root_path, image_path.strip())
    else:
        for image_path in glob(os.path.join(root_path, f"ccpd_{segment_tail}", "*.jpg")):
            yield image_path



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Implementation of gas draft."""
from configparser import ConfigParser
from textwrap import indent
from typing import Optional, Tuple

import click

from tensorbay.cli.auth import INDENT
from tensorbay.cli.tbrn import TBRN, TBRNType
from tensorbay.cli.utility import (
    ContextInfo,
    edit_message,
    error,
    exception_handler,
    format_hint,
    get_dataset_client,
)
from tensorbay.client.gas import DatasetClientType
from tensorbay.client.struct import ROOT_COMMIT_ID

_DRAFT_HINT = """
# Please enter the message for your draft.
# Lines starting with '#' will be ignored.
# And an empty draft message aborts the creation.
"""

_FULL_DRAFT_MESSAGE = """Branch:{}{}

    {}
"""


@exception_handler
def _implement_draft(  # pylint: disable=too-many-arguments
    obj: ContextInfo,
    tbrn: str,
    is_list: bool,
    edit: bool,
    close: bool,
    message: Tuple[str, ...],
) -> None:
    gas = obj.get_gas()
    tbrn_info = TBRN(tbrn=tbrn)
    dataset_client = get_dataset_client(gas, tbrn_info)

    if tbrn_info.type != TBRNType.DATASET:
        error(f'To operate a draft, "{tbrn}" must be a dataset')

    if is_list:
        _list_drafts(dataset_client, tbrn_info)
    elif edit:
        _edit_draft(dataset_client, tbrn_info, message, obj.config_parser)
    elif close:
        _close_draft(dataset_client, tbrn_info)
    else:
        _create_draft(dataset_client, tbrn_info, message, obj.config_parser)


def _create_draft(
    dataset_client: DatasetClientType,
    tbrn_info: TBRN,
    message: Tuple[str, ...],
    config_parser: ConfigParser,
) -> None:
    if tbrn_info.is_draft:
        error(f'Create a draft in draft status "{tbrn_info.get_tbrn()}" is not permitted')

    title, description = edit_message(message, _DRAFT_HINT, config_parser)
    if not title:
        error("Aborting creating draft due to empty draft message")

    dataset_client.create_draft(title=title, description=description)
    status = dataset_client.status
    draft_tbrn = TBRN(tbrn_info.dataset_name, draft_number=status.draft_number).get_colored_tbrn()
    click.echo(f'Successfully created draft "{draft_tbrn}"')
    _echo_draft(dataset_client, title, description, status.branch_name)


def _list_drafts(dataset_client: DatasetClientType, tbrn_info: TBRN) -> None:
    if tbrn_info.revision:
        error(f'list drafts based on given revision "{tbrn_info.get_tbrn()}" is not supported')

    if tbrn_info.is_draft:
        draft = dataset_client.get_draft(tbrn_info.draft_number)
        click.echo(f"Draft: {tbrn_info.get_tbrn()}")
        _echo_draft(dataset_client, draft.title, draft.description, draft.branch_name)
    else:
        for draft in dataset_client.list_drafts():
            click.echo(
                f"Draft: {TBRN(tbrn_info.dataset_name, draft_number=draft.number).get_tbrn()}"
            )
            _echo_draft(dataset_client, draft.title, draft.description, draft.branch_name)


def _echo_draft(
    dataset_client: DatasetClientType,
    title: str = "",
    description: str = "",
    branch_name: Optional[str] = None,
) -> None:
    if not branch_name:
        error("Draft should be created based on a branch.")

    branch = dataset_client.get_branch(branch_name)

    if branch.commit_id != ROOT_COMMIT_ID:
        commit_id = f"({branch.commit_id})"
    else:
        commit_id = ""

    if not title:
        title = "<no title>"
    if description:
        description = f"\n\n{indent(description, INDENT)}"
    draft_message = f"{title}{description}"
    click.echo(
        _FULL_DRAFT_MESSAGE.format(
            branch_name,
            commit_id,
            draft_message,
        )
    )


def _edit_draft(
    dataset_client: DatasetClientType,
    tbrn_info: TBRN,
    message: Tuple[str, ...],
    config_parser: ConfigParser,
) -> None:
    if not tbrn_info.is_draft:
        error("Draft number is required when editing draft")

    draft = dataset_client.get_draft()
    hint_message = format_hint(draft.title, draft.description, _DRAFT_HINT)
    title, description = edit_message(message, hint_message, config_parser)
    if not title:
        error("Aborting updating draft due to empty draft message")

    dataset_client.update_draft(title=title, description=description)
    click.echo(f'Successfully updated draft "{tbrn_info.get_colored_tbrn()}"')
    _echo_draft(dataset_client, title, description, dataset_client.status.branch_name)


def _close_draft(dataset_client: DatasetClientType, tbrn_info: TBRN) -> None:
    if not tbrn_info.draft_number:
        error("Draft number is required when editing draft")

    dataset_client._close_draft(tbrn_info.draft_number)  # pylint: disable=protected-access
    click.echo(f'Successfully closed draft "{tbrn_info.get_colored_tbrn()}"')



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Methods for open datasets of nuScenes format."""

import json
import os
from bisect import bisect
from itertools import accumulate
from typing import Any, Dict, List, Union

from tensorbay.sensor import Camera, Lidar, Radar

_SENSOR_TYPE_CLASS = {
    "lidar": Lidar,
    "radar": Radar,
    "camera": Camera,
}


def get_info_with_token(info_path: str, annotation_part: str) -> Dict[str, Any]:
    """Parse the nuScenes format label files.

    Arguments:
        info_path: The path of the JSON files.
        annotation_part: The name of JSON file.

    Returns:
        An annotation dict whose keys are tokens.

    """
    filepath = os.path.join(info_path, f"{annotation_part}.json")
    with open(filepath, "r", encoding="utf-8") as fp:
        info = json.load(fp)
    return {item.pop("token"): item for item in info}


def get_info_with_determined_token(
    info_path: str,
    annotation_part: str,
    determined_token: str = "sample_token",
    no_key_frame: bool = False,
) -> Dict[str, List[Any]]:
    """Parse the nuScenes label files with determined key.

    Arguments:
        info_path: The path of JSON files.
        annotation_part: The name of JSON file.
        determined_token: The key name of the dict.
        no_key_frame: Whether need key frame.

    Returns:
        An annotation dict whose keys are determined tokens.

    """
    filepath = os.path.join(info_path, f"{annotation_part}.json")
    with open(filepath, "r", encoding="utf-8") as fp:
        info = json.load(fp)
    info_with_keys: Dict[str, List[Any]] = {}
    for item in info:
        if no_key_frame or item["is_key_frame"]:
            info_with_keys.setdefault(item.pop(determined_token), []).append(item)
    return info_with_keys


def uncompress_rle(original_code: str) -> List[int]:
    """Uncompress function for RLE of coco API.

    Source Code: https://github.com/cocodataset/cocoapi/blob/master/common/maskApi.c#L218

    Arguments:
        original_code: The original ASCII code.

    Returns:
        The uncompressed RLE.

    """
    pointer = 0
    rle: List[int] = []
    m = 0
    while pointer < len(original_code):
        x = 0
        k = 0
        more = 1
        while more:
            char = ord(original_code[pointer]) - 48
            x |= (char & 0x1F) << 5 * k
            more = char & 0x20
            pointer += 1
            k += 1
            if not more and (char & 0x10):
                x |= -1 << 5 * k
        if m > 2:
            x += rle[m - 2]
        rle.append(x)
        m += 1
    return rle


def transpose_rle(rle: List[int], height: int, width: int) -> List[int]:
    """Transpose function for uncompressed RLE.

    Arguments:
        rle: The original RLE.
        height: The height of mask.
        width: The width of mask.

    Returns:
        RLE which has been transposed.

    """
    accumulate_values = list(accumulate(rle))
    flag = 0
    count = 0
    rle = []
    for i in range(height):
        for index in range(i, i + height * width, height):
            if bisect(accumulate_values, index) % 2 == flag:
                count += 1
                continue
            flag = 1 - flag
            rle.append(count)
            count = 1
    rle.append(count)
    return rle


def get_sensor(
    sensor_type: str,
    sensor_name: str,
    calibrated_sensor_info: Dict[str, Any],
) -> Union[Camera, Lidar, Radar]:
    """Load the sensor information for the fusion segment.

    Arguments:
        sensor_type: The type of the sensor.
        sensor_name: The name of the sensor.
        calibrated_sensor_info: The information of the sensor.

    Returns:
        The sensor information of the fusion segment.

    """
    sensor = _SENSOR_TYPE_CLASS[sensor_type](sensor_name)
    sensor.set_extrinsics(
        translation=calibrated_sensor_info["translation"],
        rotation=calibrated_sensor_info["rotation"],
    )
    intrinsics = calibrated_sensor_info["camera_intrinsic"]
    if intrinsics:
        sensor.set_camera_matrix(matrix=intrinsics)  # type: ignore[attr-defined]
        distortion = calibrated_sensor_info.get("camera_distortion", None)
        if distortion:
            distortion_keys = ("k1", "k2", "p1", "p2", "k3", "k4")
            kwargs = dict(zip(distortion_keys, distortion))
            sensor.set_distortion_coefficients(**kwargs)  # type: ignore[attr-defined]
    return sensor  # type: ignore[return-value]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=import-outside-toplevel

"""Customization of the click classes for TensorBay CLI."""

from types import MethodType
from typing import Any, Optional, Sequence, Tuple, Union

import click


class CustomCommand(click.Command):
    """Wrapper class of ``click.Command`` for CLI commands with custom help.

    Arguments:
        kwargs: The keyword arguments pass to ``click.Command``.

    """

    def __init__(self, **kwargs: Any) -> None:
        self.synopsis = kwargs.pop("synopsis", ())
        super().__init__(**kwargs)

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Writes the custom help into the formatter if it exists.

        Override the original ``click.Command.format_help`` method by
        adding :meth:`CustomCommand.format_synopsis` to form custom help message.

        Arguments:
            ctx: The context of the command.
            formatter: The help formatter of the command.

        """
        formatter.width = 100
        self.format_usage(ctx, formatter)
        self.format_help_text(ctx, formatter)
        self.format_synopsis(formatter)  # Add synopsis in command help.
        self.format_options(ctx, formatter)
        self.format_epilog(ctx, formatter)

    def format_synopsis(self, formatter: click.HelpFormatter) -> None:
        """Wirte the synopsis to the formatter if exist.

        Arguments:
            formatter: The help formatter of the command.

        """
        if not self.synopsis:
            return

        with formatter.section("Synopsis"):
            for example in self.synopsis:
                formatter.write_text(example)


class DeprecatedCommand(CustomCommand):
    """Customized ``click.Command`` wrapper class for deprecated CLI commands.

    Arguments:
        args: The positional arguments pass to ``click.Command``.
        since: The version the function is deprecated.
        removed_in: The version the function will be removed in.
        substitute: The substitute command.
        kwargs: The keyword arguments pass to ``click.Command``.
    """

    def __init__(
        self,
        *args: Any,
        since: str,
        removed_in: Optional[str] = None,
        substitute: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        messages = [
            f'DeprecationWarning: The command "{self.name}" is deprecated since version {since}.'
        ]
        if removed_in:
            messages.append(f'It will be removed in version "{removed_in}".')
        if substitute:
            messages.append(f'Use "{substitute}" instead.')

        self.deprecated_message = " ".join(messages)

    def invoke(self, ctx: click.Context) -> Any:
        """This invokes the command to print the deprecated message.

        Arguments:
            ctx: The click Context.

        Returns:
            The invoke result of ``click.Command``.

        """
        click.secho(self.deprecated_message, fg="red", err=True)
        return super().invoke(ctx)


class DeprecatedOption(click.Option):
    """Customize deprecated option."""

    def __init__(
        self,
        *args: Any,
        deprecated: Union[str, Tuple[str, ...]],
        since: str,
        removed_in: Optional[str] = None,
        preferred: Optional[Union[str, Tuple[str, ...]]] = None,
        **kwargs: Any,
    ) -> None:
        if isinstance(deprecated, str):
            self.deprecated: Tuple[str, ...] = (deprecated,)
        else:
            self.deprecated = deprecated
        self.since = since
        self.removed_in = removed_in
        self.preferred = preferred
        super().__init__(*args, **kwargs)

    def join_options(self, options: Sequence[str]) -> Tuple[str, bool]:
        """Join the options that are not deprecated.

        Copied from ``click.formatting.join_options``
        and add some code to exclude the deprecated opts in help message.

        Arguments:
            options: The options.

        Returns:
            The joined option string and whether there is slash prefix in options.

        """
        prefix_opt = []
        any_prefix_is_slash = False

        for opt in options:
            if opt in self.deprecated:  # Exclude the deprecated opts in help message.
                continue

            prefix = click.parser.split_opt(opt)[0]

            if prefix == "/":
                any_prefix_is_slash = True

            prefix_opt.append((len(prefix), opt))

        prefix_opt.sort(key=lambda x: x[0])
        return ", ".join(x[1] for x in prefix_opt), any_prefix_is_slash

    def get_help_record(self, ctx: click.Context) -> Optional[Tuple[str, str]]:
        """Get help record.

        Arguments:
            ctx: The context of the option.

        Returns:
            The option help message.

        """
        help_record = super().get_help_record(ctx)
        if help_record is None:
            return None

        any_prefix_is_slash = False

        def _write_opts(opts: Sequence[str]) -> str:
            nonlocal any_prefix_is_slash

            # Use `self.join_options` instead of `click.formatting.join_options`
            # to exclude the deprecated opts in help message.
            opt_message, any_slashes = self.join_options(opts)

            if any_slashes:
                any_prefix_is_slash = True

            if not self.is_flag and not self.count:
                opt_message += f" {self.make_metavar()}"

            return opt_message

        opt_messages = [_write_opts(self.opts)]

        if self.secondary_opts:
            opt_messages.append(_write_opts(self.secondary_opts))

        return ("; " if any_prefix_is_slash else " / ").join(opt_messages), help_record[1]


class DeprecatedOptionsCommand(CustomCommand):
    """Customize command with deprecated options."""

    @staticmethod
    def _process(
        option: click.parser.Option,
        value: Any,
        state: click.parser.ParsingState,
    ) -> None:
        option_obj: DeprecatedOption = option.obj  # type: ignore[assignment]

        # reach up the stack and get 'opt'
        import inspect

        frame = inspect.currentframe()
        if frame and frame.f_back:
            opt = frame.f_back.f_locals.get("opt")
        else:
            opt = None

        if opt in option_obj.deprecated:
            messages = [
                "DeprecationWarning: "
                f'The option "{opt}" is deprecated since version {option_obj.since}.'
            ]
            if option_obj.removed_in:
                messages.append(f"It will be removed in version {option_obj.removed_in}.")
            if option_obj.preferred:
                messages.append(f'Please use "{option_obj.preferred}" instead.')
            click.secho(" ".join(messages), fg="yellow")
        return click.parser.Option.process(option, value, state)

    def make_parser(self, ctx: click.Context) -> click.OptionParser:
        """Hook 'make_parser' and check whether the name used to invoke the option is preferred.

        Arguments:
            ctx: The context of the command.

        Returns:
            The option parser.

        """
        parser = super().make_parser(ctx)

        options = set(parser._short_opt.values()) | set(  # pylint: disable=protected-access
            parser._long_opt.values()  # pylint: disable=protected-access
        )

        for option in options:
            if not isinstance(option.obj, DeprecatedOption):
                continue

            option.process = MethodType(self._process, option)  # type: ignore[assignment]
        return parser



#!/usr/bin/env python3
#
# Copytright 2020 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of UAVDT."""

from tensorbay.opendataset.UAVDT.loader import UAVDT

__all__ = ["UAVDT"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of AnimalPose5 dataset and AnimalPose7 dataset."""

import json
import os
from typing import Any, Iterable, Iterator, List, Tuple

from tensorbay.dataset import Data, Dataset
from tensorbay.exception import ModuleImportError
from tensorbay.geometry import Keypoint2D
from tensorbay.label import LabeledBox2D, LabeledKeypoints2D
from tensorbay.opendataset._utility import glob

DATASET_NAME_5 = "AnimalPose5"
DATASET_NAME_7 = "AnimalPose7"

_KEYPOINT_TO_INDEX = {
    "L_Eye": 0,
    "R_Eye": 1,
    "Nose": 4,
    "L_EarBase": 2,
    "R_EarBase": 3,
    "Throat": 5,
    "L_F_Elbow": 8,
    "R_F_Elbow": 9,
    "L_F_Paw": 16,
    "R_F_Paw": 17,
    "Withers": 7,
    "TailBase": 6,
    "L_B_Paw": 18,
    "R_B_Paw": 19,
    "L_B_Elbow": 10,
    "R_B_Elbow": 11,
    "L_F_Knee": 12,
    "R_F_Knee": 13,
    "L_B_Knee": 14,
    "R_B_Knee": 15,
}


def AnimalPose5(path: str) -> Dataset:
    """`5 Categories Animal-Pose <https://sites.google.com/view/animal-pose/>`_ dataset.

    The file structure should be like::

        <path>
            keypoint_image_part1/
                cat/
                    2007_000549.jpg
                    2007_000876.jpg
                    ...
                ...
            PASCAL2011_animal_annotation/
                cat/
                    2007_000549_1.xml
                    2007_000876_1.xml
                    2007_000876_2.xml
                    ...
                ...
            animalpose_image_part2/
                cat/
                    ca1.jpeg
                    ca2.jpeg
                    ...
                ...
            animalpose_anno2/
                cat/
                    ca1.xml
                    ca2.xml
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME_5)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_5.json"))
    animals = dataset.catalog.keypoints2d.categories.keys()

    for segment_name, _get_data in _DATA_GETTERS.items():
        segment = dataset.create_segment(segment_name)
        for data in _get_data(root_path, animals):
            segment.append(data)

    return dataset


def _get_data_part1(root_path: str, aniamls: Iterable[str]) -> Iterator[Data]:
    try:
        import xmltodict  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(module_name=error.name) from error

    for animal in aniamls:
        for image_path in glob(os.path.join(root_path, "keypoint_image_part1", animal, "*.jpg")):
            data = Data(image_path, target_remote_path=f"{animal}/{os.path.basename(image_path)}")

            for annotation_path in glob(
                os.path.join(
                    root_path,
                    "PASCAL2011_animal_annotation",
                    animal,
                    f"{os.path.splitext(os.path.basename(image_path))[0]}_*.xml",
                )
            ):

                with open(annotation_path, encoding="utf-8") as fp:
                    labels: Any = xmltodict.parse(fp.read())

                box2d = labels["annotation"]["visible_bounds"]
                data.label.box2d = [
                    LabeledBox2D.from_xywh(
                        x=float(box2d["@xmin"]),
                        y=float(box2d["@ymin"]),
                        width=float(box2d["@width"]),
                        height=float(box2d["@height"]),
                        category=animal,
                    )
                ]

                keypoints2d: List[Tuple[float, float, int]] = [()] * 20  # type: ignore[list-item]
                for keypoint in labels["annotation"]["keypoints"]["keypoint"]:
                    keypoints2d[_KEYPOINT_TO_INDEX[keypoint["@name"]]] = (
                        float(keypoint["@x"]),
                        float(keypoint["@y"]),
                        int(keypoint["@visible"]),
                    )
                data.label.keypoints2d = [LabeledKeypoints2D(keypoints2d, category=animal)]

            yield data


def _get_data_part2(root_path: str, aniamls: Iterable[str]) -> Iterator[Data]:
    try:
        import xmltodict  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(module_name=error.name) from error

    for animal in aniamls:
        for image_path in glob(os.path.join(root_path, "animalpose_image_part2", animal, "*.jpeg")):
            data = Data(image_path, target_remote_path=f"{animal}/{os.path.basename(image_path)}")

            annotation_path = os.path.join(
                root_path,
                "animalpose_anno2",
                animal,
                f"{os.path.splitext(os.path.basename(image_path))[0]}.xml",
            )

            with open(annotation_path, encoding="utf-8") as fp:
                labels: Any = xmltodict.parse(fp.read())

            box2d = labels["annotation"]["visible_bounds"]
            data.label.box2d = [
                LabeledBox2D.from_xywh(
                    x=float(box2d["@xmin"]),
                    y=float(box2d["@xmax"]),  # xmax means ymin in the annotation
                    width=float(box2d["@width"]),
                    height=float(box2d["@height"]),
                    category=animal,
                )
            ]

            keypoints2d = LabeledKeypoints2D(category=animal)
            for keypoint in labels["annotation"]["keypoints"]["keypoint"]:
                keypoints2d.append(
                    Keypoint2D(
                        float(keypoint["@x"]), float(keypoint["@y"]), int(keypoint["@visible"])
                    )
                )
            data.label.keypoints2d = [keypoints2d]
            yield data


_DATA_GETTERS = {"part1": _get_data_part1, "part2": _get_data_part2}


def AnimalPose7(path: str) -> Dataset:
    """`7 Categories Animal-Pose <https://sites.google.com/view/animal-pose/>`_ dataset.

    The file structure should be like::

        <path>
            bndbox_image/
                antelope/
                    Img-77.jpg
                    ...
                ...
            bndbox_anno/
                antelope.json
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        loaded :class:`~tensorbay.dataset.dataset.Dataset` object.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME_7)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_7.json"))

    segment = dataset.create_segment()

    for animal in dataset.catalog.box2d.categories.keys():
        with open(os.path.join(root_path, "bndbox_anno", f"{animal}.json"), encoding="utf-8") as fp:
            annotations = json.load(fp)
        for image_name, box2ds in annotations.items():
            image_path = os.path.join(root_path, "bndbox_image", animal, image_name)
            data = Data(image_path, target_remote_path=f"{animal}/{image_name}")
            data.label.box2d = []

            for box2d in box2ds:
                coordinates = box2d["bndbox"]
                data.label.box2d.append(
                    LabeledBox2D(
                        float(coordinates["xmin"]),
                        float(coordinates["ymin"]),
                        float(coordinates["xmax"]),
                        float(coordinates["ymax"]),
                        category=animal,
                    )
                )

            segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the The Car Connection Picture dataset."""

from tensorbay.opendataset.CarConnection.loader import CarConnection

__all__ = ["CarConnection"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of Flower17 dataset and Flower102 dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.exception import ModuleImportError
from tensorbay.label import Classification

DATASET_NAME_17 = "Flower17"
DATASET_NAME_102 = "Flower102"
_SEGMENT_NAMES_17 = {"train": "trn1", "validation": "val1", "test": "tst1"}
_SEGMENT_NAMES_102 = {"train": "trnid", "validation": "valid", "test": "tstid"}


def Flower17(path: str) -> Dataset:
    """`17 Category Flower <http://www.robots.ox.ac.uk/~vgg/data/flowers/17/index.html>`_ dataset.

    The dataset are 3 separate splits.
    The results in the paper are averaged over the 3 splits.
    We just use (trn1, val1, tst1) to split it.

    The file structure should be like::

                <path>
                    jpg/
                        image_0001.jpg
                        ...
                    datasplits.mat

    Arguments:
        path: The root directory of the dataset.

    Raises:
        ModuleImportError: When the module "scipy" can not be found.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    try:
        from scipy.io import loadmat  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(module_name=error.name) from error

    root_path = os.path.abspath(os.path.expanduser(path))
    segment_info = loadmat(os.path.join(root_path, "datasplits.mat"))

    dataset = Dataset(DATASET_NAME_17)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_17.json"))
    index_to_category = dataset.catalog.classification.get_index_to_category()
    for key, value in _SEGMENT_NAMES_17.items():
        segment = dataset.create_segment(key)
        segment_info[value][0].sort()
        for index in segment_info[value][0]:
            data = Data(os.path.join(root_path, "jpg", f"image_{index:04d}.jpg"))

            # There are 80 images for each category
            data.label.classification = Classification(
                category=index_to_category[(index - 1) // 80]
            )
            segment.append(data)

    return dataset


def Flower102(path: str) -> Dataset:
    """`102 Category Flower <http://www.robots.ox.ac.uk/~vgg/data/flowers/102/index.html>`_ dataset.

    The file structure should be like::

        <path>
            jpg/
                image_00001.jpg
                ...
            imagelabels.mat
            setid.mat

    Arguments:
        path: The root directory of the dataset.

    Raises:
        ModuleImportError: When the module "scipy" can not be found.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    try:
        from scipy.io import loadmat  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(module_name=error.name) from error

    root_path = os.path.abspath(os.path.expanduser(path))
    labels = loadmat(os.path.join(root_path, "imagelabels.mat"))["labels"][0]
    segment_info = loadmat(os.path.join(root_path, "setid.mat"))

    dataset = Dataset(DATASET_NAME_102)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_102.json"))
    index_to_category = dataset.catalog.classification.get_index_to_category()
    for key, value in _SEGMENT_NAMES_102.items():
        segment = dataset.create_segment(key)
        segment_info[value][0].sort()
        for index in segment_info[value][0]:
            data = Data(os.path.join(root_path, "jpg", f"image_{index:05d}.jpg"))
            data.label.classification = Classification(
                index_to_category[int(labels[index - 1]) - 1]
            )
            segment.append(data)
    return dataset



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of RP2K dataset."""

import os
from glob import glob
from typing import Iterable, List

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification

DATASET_NAME = "RP2K"


def RP2K(path: str) -> Dataset:
    """`RP2K <https://www.pinlandata.com/rp2k_dataset>`_ dataset.

    The file structure of RP2K looks like::

        <path>
            all/
                test/
                    <catagory>/
                        <image_name>.jpg
                        ...
                    ...
                train/
                    <catagory>/
                        <image_name>.jpg
                        ...
                    ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.join(os.path.abspath(os.path.expanduser(path)), "all")
    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name in ("train", "test"):
        segment = dataset.create_segment(segment_name)
        segment_path = os.path.join(root_path, segment_name)
        categories = os.listdir(segment_path)
        categories.sort()
        for category in categories:
            category_dir = os.path.join(segment_path, category)
            if not os.path.isdir(category_dir):
                continue
            image_paths = _glob(category_dir, ("*.jpg", "*.png"))
            for image_path in image_paths:
                remote_path = os.path.basename(image_path).replace(" ", "_")
                data = Data(local_path=image_path, target_remote_path=remote_path)
                data.label.classification = Classification(category)
                segment.append(data)

    return dataset


def _glob(category_dir: str, patterns: Iterable[str]) -> List[str]:
    file_paths = []

    for pattern in patterns:
        file_paths.extend(glob(os.path.join(category_dir, pattern)))

    file_paths.sort()
    return file_paths



#!/usr/bin/env python3
#
# Copytright 2020 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of UAVDT dataset."""

import csv
import os
from collections import defaultdict
from typing import Dict, List

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Box2DSubcatalog, Classification, ClassificationSubcatalog, LabeledBox2D
from tensorbay.opendataset._utility import glob

_DATASET_NAME = "UAVDT"

_MODES = ("train", "test")

_FIELDNAMES = (
    "frame_index",
    "target_id",
    "bbox_left",
    "bbox_top",
    "bbox_width",
    "bbox_height",
    "out_of_view",
    "occlusion",
    "object_category",
)

_XYWH_KEYS = _FIELDNAMES[2:6]


def UAVDT(path: str) -> Dataset:
    """`UAVDT <https://paperswithcode.com/dataset/uavdt>`_ dataset.

    The "score", "in-view", "occlusion" fields in MOT Groundtruth file(``*_gt.txt``) are constant,
    and other fields in that file are the same with such fields in DET Groundtruth file
    (``*_gt_whole.txt``). Therefore, they are not included in the dataloader.

    The Ignore Areas file(``*_gt_ignore.txt``) is useless,
    so they are not included in the dataloader neither.

    The file structure of UAVDT looks like::

        <path>
            M_attr/
                test/
                    M0203_attr.txt
                    ...
                train/
                    M0101_attr.txt
                    ...
            UAVDT_Benchmark_M/
                M0101/
                    img000001.jpg
                    ...
                ...
            UAV-benchmark-MOTD_v1.0/
                GT/
                    M0101_gt_ignore.txt
                    M0101_gt.txt
                    M0101_gt_whole.txt
                    ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(_DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for mode in _MODES:
        for sequence_attributes_path in glob(os.path.join(root_path, "M_attr", mode, "*.txt")):
            sequence = os.path.basename(sequence_attributes_path)[:5]
            segment = dataset.create_segment(f"{mode}-{sequence}")

            classification = _extract_classification(
                sequence_attributes_path, dataset.catalog.classification
            )

            frame_id_box2d_map = _extract_box2d(root_path, sequence, dataset.catalog.box2d)

            image_paths = glob(os.path.join(root_path, "UAV-benchmark-M", sequence, "*.jpg"))
            for image_path in image_paths:
                data = Data(image_path)
                data.label.classification = classification
                # The image_name looks like `img<frame_id>.jpg`.
                # The frame_id is consist of six digital numbers.
                data.label.box2d = frame_id_box2d_map[int(os.path.basename(image_path)[3:9])]
                segment.append(data)

    return dataset


def _extract_classification(
    path: str, classification_catalog: ClassificationSubcatalog
) -> Classification:
    with open(path, encoding="utf-8") as fp:
        attribute_names = classification_catalog.attributes.keys()
        csv_reader = csv.reader(fp)
        elements = next(csv_reader)
        attributes = {
            attribute_name: bool(int(value))
            for attribute_name, value in zip(attribute_names, elements)
        }

    return Classification(attributes=attributes)


def _extract_box2d(
    path: str, sequence: str, box2d_catalog: Box2DSubcatalog
) -> Dict[int, List[LabeledBox2D]]:
    attributes = box2d_catalog.attributes
    category_names = box2d_catalog.categories.keys()

    out_of_view_level = attributes["out_of_view"].enum
    occlusion_level = attributes["occlusion"].enum

    ground_truth_path = os.path.join(path, "UAV-benchmark-MOTD_v1.0", "GT")
    frame_id_ground_truth_map = defaultdict(list)
    with open(os.path.join(ground_truth_path, f"{sequence}_gt_whole.txt"), encoding="utf-8") as fp:
        for elements in csv.DictReader(fp, fieldnames=_FIELDNAMES):
            box2d = LabeledBox2D.from_xywh(
                *(int(elements[key]) for key in _XYWH_KEYS),
                category=category_names[int(elements["object_category"]) - 1],
                attributes={
                    "out_of_view": out_of_view_level[int(elements["out_of_view"]) - 1],
                    "occlusion": occlusion_level[int(elements["occlusion"]) - 1],
                },
                instance=elements["target_id"],
            )
            frame_id = int(elements["frame_index"])
            frame_id_ground_truth_map[frame_id].append(box2d)

    return frame_id_ground_truth_map



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the COCO2017 dataset."""

from tensorbay.opendataset.COCO2017.loader import COCO2017

__all__ = ["COCO2017"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""TensorBay cutoms exceptions.

The class hierarchy for TensorBay custom exceptions is::

     +-- TensorBayException
         +-- ClientError
             +-- StatusError
             +-- DatasetTypeError
             +-- FrameError
             +-- ResponseError
                 +-- AccessDeniedError
                 +-- ForbiddenError
                 +-- InvalidParamsError
                 +-- NameConflictError
                 +-- RequestParamsMissingError
                 +-- ResourceNotExistError
                 +-- InternalServerError
                 +-- UnauthorizedError
        +-- UtilityError
            +-- AttrError
         +-- TBRNError
         +-- OpenDatasetError
             +-- NoFileError
             +-- FileStructureError

:class:`OperationError` is removed in version v1.13.0.
Use :class:`StatusError` or :class:`ValueError` instead.

"""

from typing import Dict, Optional, Type, Union

from requests.models import Response


class TensorBayException(Exception):
    """This is the base class for TensorBay custom exceptions.

    Arguments:
       message: The error message.

    """

    def __init__(self, message: Optional[str] = None):
        super().__init__()
        self._message = message

    def __str__(self) -> str:
        return self._message if self._message else ""


class ClientError(TensorBayException):
    """This is the base class for custom exceptions in TensorBay client module."""


class StatusError(ClientError):
    """This class defines the exception for illegal status.

    Arguments:
        is_draft: Whether the status is draft.
        message: The error message.

    """

    def __init__(self, message: Optional[str] = None, *, is_draft: Optional[bool] = None) -> None:
        super().__init__()
        if is_draft is None:
            self._message = message
        else:
            required_status = "commit" if is_draft else "draft"
            self._message = f"The status is not {required_status}"


class DatasetTypeError(ClientError):
    """This class defines the exception for incorrect type of the requested dataset.

    Arguments:
        dataset_name: The name of the dataset whose requested type is wrong.
        is_fusion: Whether the dataset is a fusion dataset.

    """

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        dataset_name: Optional[str] = None,
        is_fusion: Optional[bool] = None,
    ) -> None:
        super().__init__(message)
        self._dataset_name = dataset_name
        self._is_fusion = is_fusion

    def __str__(self) -> str:
        if self._dataset_name and self._is_fusion:
            return (
                f'Dataset "{self._dataset_name}" is '
                f'{"" if self._is_fusion else "not "}a fusion dataset'
            )
        return super().__str__()


class FrameError(ClientError):
    """This class defines the exception for incorrect frame id."""


class ResponseError(ClientError):
    """This class defines the exception for post response error.

    Arguments:
        response: The response of the request.

    Attributes:
        response: The response of the request.

    """

    # https://github.com/python/mypy/issues/6473
    _INDENT = " " * len(__qualname__)  # type: ignore[name-defined]

    STATUS_CODE: int

    def __init__(
        self, message: Optional[str] = None, *, response: Optional[Response] = None
    ) -> None:
        super().__init__(message)
        if response is not None:
            self.response = response

    def __init_subclass__(cls) -> None:
        cls._INDENT = " " * len(cls.__name__)

    def __str__(self) -> str:
        if hasattr(self, "response"):
            return (
                f"Unexpected status code({self.response.status_code})! {self.response.url}!"
                f"\n{self._INDENT}  {self.response.text}"
            )
        return super().__str__()


class AccessDeniedError(ResponseError):
    """This class defines the exception for access denied response error."""

    STATUS_CODE = 403


class ForbiddenError(ResponseError):
    """This class defines the exception for illegal operations Tensorbay forbids."""

    STATUS_CODE = 403


class InvalidParamsError(ResponseError):
    """This class defines the exception for invalid parameters response error.

    Arguments:
        response: The response of the request.
        param_name: The name of the invalid parameter.
        param_value: The value of the invalid parameter.

    Attributes:
        response: The response of the request.

    """

    STATUS_CODE = 400

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        response: Optional[Response] = None,
        param_name: Optional[str] = None,
        param_value: Optional[str] = None,
    ) -> None:
        super().__init__(message, response=response)
        self._param_name = param_name
        self._param_value = param_value

    def __str__(self) -> str:
        if self._param_name and self._param_value:
            messages = [f"Invalid {self._param_name}: {self._param_value}."]
            if self._param_name == "path":
                messages.append("Remote path should follow linux style.")

            return f"\n{self._INDENT}".join(messages)
        return super().__str__()


class NameConflictError(ResponseError):
    """This class defines the exception for name conflict response error.

    Arguments:
        response: The response of the request.
        resource: The type of the conflict resource.
        identification: The identification of the conflict resource.

    Attributes:
        response: The response of the request.

    """

    STATUS_CODE = 409

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        response: Optional[Response] = None,
        resource: Optional[str] = None,
        identification: Union[int, str, None] = None,
    ) -> None:
        super().__init__(message, response=response)
        self._resource = resource
        self._identification = identification

    def __str__(self) -> str:
        if self._resource and self._identification:
            return f"The {self._resource}: {self._identification} already exists."
        return super().__str__()


class RequestParamsMissingError(ResponseError):
    """This class defines the exception for request parameters missing response error."""

    STATUS_CODE = 400


class ResourceNotExistError(ResponseError):
    """This class defines the exception for resource not existing response error.

    Arguments:
        response: The response of the request.
        resource: The type of the conflict resource.
        identification: The identification of the conflict resource.

    Arguments:
        response: The response of the request.

    """

    STATUS_CODE = 404

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        response: Optional[Response] = None,
        resource: Optional[str] = None,
        identification: Union[int, str, None] = None,
    ) -> None:
        super().__init__(message, response=response)
        self._resource = resource
        self._identification = identification

    def __str__(self) -> str:
        if self._resource and self._identification:
            return f"The {self._resource}: {self._identification} does not exist."
        return super().__str__()


class InternalServerError(ResponseError):
    """This class defines the exception for internal server error."""

    STATUS_CODE = 500


class UnauthorizedError(ResponseError):
    """This class defines the exception for unauthorized response error."""

    STATUS_CODE = 401


class OpenDatasetError(TensorBayException):
    """This is the base class for custom exceptions in TensorBay opendataset module."""


class NoFileError(OpenDatasetError):
    """This class defines the exception for no matching file found in the opendataset directory.

    Arguments:
        pattern: Glob pattern.

    """

    def __init__(self, message: Optional[str] = None, *, pattern: Optional[str] = None) -> None:
        super().__init__(message)
        self._pattern = pattern

    def __str__(self) -> str:
        if self._pattern:
            return f'No file follows the giving pattern "{self._pattern}"'
        return super().__str__()


class FileStructureError(OpenDatasetError):
    """This class defines the exception for incorrect file structure in opendataset directory."""


class ModuleImportError(OpenDatasetError, ModuleNotFoundError):
    """This class defines the exception for import error of optional module in opendataset module.

    Arguments:
        module_name: The name of the optional module.
        package_name: The package name of the optional module.

    """

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        module_name: Optional[str] = None,
        package_name: Optional[str] = None,
    ) -> None:
        super().__init__(message)
        self._module_name = module_name
        self._package_name = package_name if package_name else module_name

    def __str__(self) -> str:
        if self._module_name:
            return (
                f"No module named {self._module_name}."
                "\n"
                f'\n    To install the module, please run: "pip3 install {self._package_name}"'
                "\n"
            )
        return super().__str__()


class TBRNError(TensorBayException):
    """This class defines the exception for invalid TBRN."""


ResponseErrorDistributor: Dict[str, Type[ResponseError]] = {
    "AccessDenied": AccessDeniedError,
    "Forbidden": ForbiddenError,
    "InvalidParamsValue": InvalidParamsError,
    "NameConflict": NameConflictError,
    "RequestParamsMissing": RequestParamsMissingError,
    "ResourceNotExist": ResourceNotExistError,
    "InternalServerError": InternalServerError,
    "Unauthorized": UnauthorizedError,
}


class UtilityError(TensorBayException):
    """This is the base class for custom exceptions in TensorBay utility module."""


class AttrError(UtilityError):
    """This class defines the exception for dynamic attr have default value."""

    def __str__(self) -> str:
        return "Dynamic attr cannot have default value."



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name
# pylint: disable=line-too-long

"""Dataloader of CarConnectionPictureD dataset."""

import os
from typing import Tuple, Union

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification
from tensorbay.opendataset._utility import glob

DATASET_NAME = "CarConnectionPicture"


def CarConnection(path: str) -> Dataset:
    """`Car Connection Picture <https://github.com/nicolas-gervais\
    /predicting-car-price-from-scraped-data/tree/master/picture-scraper>`_ dataset.

    The file structure should be like::

        <path>
            <imagename>.jpg
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    image_paths = glob(os.path.join(root_path, "*.jpg"))
    keys = dataset.catalog.classification.attributes.keys()

    for image_path in image_paths:
        data = Data(image_path)
        basename = os.path.basename(image_path)
        label = _extract_label_from_basename(keys, basename)
        data.label.classification = label
        segment.append(data)

    return dataset


def _transfer_attribute_type(item: str) -> Union[int, str, None]:
    if item == "nan":
        return None
    if item.isnumeric():
        return int(item)

    return item


def _extract_label_from_basename(keys: Tuple[str, ...], filename: str) -> Classification:
    make, model, *spec_values = filename.split("_")[:-1]

    attributes = dict(zip(keys, map(_transfer_attribute_type, spec_values)))

    category = ".".join((make, model))

    return Classification(attributes=attributes, category=category)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of DownsampledImagenet dataset."""

import os

from tensorbay.dataset import Data, Dataset, Segment
from tensorbay.opendataset._utility import glob

DATASET_NAME = "DownsampledImagenet"
SEGMENT_NAMES = ["train_32x32", "train_64x64", "valid_32x32", "valid_64x64"]


def DownsampledImagenet(path: str) -> Dataset:
    """`Downsampled Imagenet <https://www.tensorflow.org/datasets\
    /catalog/downsampled_imagenet>`_ dataset.

    The file structure should be like::

        <path>
            valid_32x32/
                <imagename>.png
                ...
            valid_64x64/
                <imagename>.png
                ...
            train_32x32/
                <imagename>.png
                ...
            train_64x64/
                <imagename>.png
                ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)

    for segment_name in SEGMENT_NAMES:
        dataset.add_segment(_get_segment(segment_name, root_path))
    return dataset


def _get_segment(path: str, segment_name: str) -> Segment:
    segment = Segment(segment_name)
    image_paths = glob(os.path.join(path, segment_name, "*.png"))

    for image_path in image_paths:
        segment.append(Data(image_path))
    return segment



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Utility classes."""

from tensorbay.utility.attr import AttrsMixin, attr, attr_base, camel, upper
from tensorbay.utility.common import EqMixin, MatrixType, common_loads, locked
from tensorbay.utility.deprecated import (
    DefaultValueDeprecated,
    Deprecated,
    Disable,
    KwargsDeprecated,
)
from tensorbay.utility.file import URL, FileMixin, RemoteFileMixin
from tensorbay.utility.itertools import chunked
from tensorbay.utility.name import NameList, NameMixin, SortedNameList
from tensorbay.utility.repr import ReprMixin, ReprType, repr_config
from tensorbay.utility.requests import Tqdm, UserResponse, UserSession, config, get_session
from tensorbay.utility.type import TypeEnum, TypeMixin, TypeRegister
from tensorbay.utility.user import (
    UserMapping,
    UserMutableMapping,
    UserMutableSequence,
    UserSequence,
)

__all__ = [
    "AttrsMixin",
    "DefaultValueDeprecated",
    "Deprecated",
    "Disable",
    "EqMixin",
    "FileMixin",
    "KwargsDeprecated",
    "MatrixType",
    "NameList",
    "NameMixin",
    "RemoteFileMixin",
    "ReprMixin",
    "ReprType",
    "SortedNameList",
    "Tqdm",
    "TypeEnum",
    "TypeMixin",
    "TypeRegister",
    "URL",
    "UserMapping",
    "UserMutableMapping",
    "UserMutableSequence",
    "UserResponse",
    "UserSequence",
    "UserSession",
    "attr",
    "attr_base",
    "camel",
    "chunked",
    "common_loads",
    "config",
    "locked",
    "repr_config",
    "get_session",
    "upper",
]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Segment and FusionSegment.

Segment is a concept in :class:`~tensorbay.dataset.dataset.Dataset`.
It is the structure that composes :class:`~tensorbay.dataset.dataset.Dataset`,
and consists of a series of :class:`~tensorbay.dataset.data.Data` without sensor information.

Fusion segment is a concept in :class:`~tensorbay.dataset.dataset.FusionDataset`.
It is the structure that composes :class:`~tensorbay.dataset.dataset.FusionDataset`,
and consists of a list of :class:`~tensorbay.dataset.frame.Frame`
along with multiple :class:`~tensorbay.sensor.sensor.Sensors`.

"""

from typing import TYPE_CHECKING, Any, Callable, MutableSequence, Optional, Type, TypeVar

from tensorbay.dataset.frame import Frame
from tensorbay.sensor import Sensors
from tensorbay.utility import NameMixin, ReprType, UserMutableSequence

if TYPE_CHECKING:
    from tensorbay.client.dataset import DatasetClient, FusionDatasetClient
    from tensorbay.client.segment import FusionSegmentClient, SegmentClient
    from tensorbay.dataset.data import DataBase

_S = TypeVar("_S", bound="Segment")
_FS = TypeVar("_FS", bound="FusionSegment")


class Segment(NameMixin, UserMutableSequence["DataBase._Type"]):
    """This class defines the concept of segment.

    Segment is a concept in :class:`~tensorbay.dataset.dataset.Dataset`.
    It is the structure that composes :class:`~tensorbay.dataset.dataset.Dataset`,
    and consists of a series of :class:`~tensorbay.dataset.data.Data` without sensor information.

    If the segment is inside of a time-continuous :class:`~tensorbay.dataset.dataset.Dataset`,
    the time continuity of the data should be indicated by
    :meth`~graviti.dataset.data.Data.remote_path`.

    Since :class:`Segment` extends :class:`~tensorbay.utility.user.UserMutableSequence`,
    its basic operations are the same as a list's.

    To initialize a Segment and add a :class:`~tensorbay.dataset.data.Data` to it:

    .. code:: python

        segment = Segment(segment_name)
        segment.append(Data())

    Arguments:
        name: The name of the segment, whose default value is an empty string.
        client: The DatasetClient if you want to read the segment from tensorbay.

    """

    _repr_type = ReprType.SEQUENCE

    def __init__(self, name: str = "default", client: Optional["DatasetClient"] = None) -> None:
        super().__init__(name)

        if client:
            self._client = client.get_segment(name)
            self._data = self._client.list_data()  # type: ignore[assignment]
            self._repr_non_empty = True
        else:
            self._data = []

    @classmethod
    def _from_client(cls: Type[_S], client: "SegmentClient") -> _S:
        """Init a Segment from :class:`~tensorbay.client.segment.SegmentClient`.

        Arguments:
            client: The :class:`~tensorbay.client.segment.SegmentClient`.

        Returns:
            The Segment of the input :class:`~tensorbay.client.segment.SegmentClient`.

        """
        segment = cls(client.name)
        segment._client = client
        segment._data = client.list_data()  # type: ignore[assignment]
        segment._repr_non_empty = True
        return segment

    def sort(
        self,
        *,
        key: Callable[["DataBase._Type"], Any] = lambda data: data.path,
        reverse: bool = False,
    ) -> None:
        """Sort the list in ascending order and return None.

        The sort is in-place (i.e. the list itself is modified) and stable (i.e. the
        order of two equal elements is maintained).

        Arguments:
            key: If a key function is given, apply it once to each item of the segment,
                and sort them according to their function values in ascending or descending order.
                By default, the data within the segment is sorted by fileuri.
            reverse: The reverse flag can be set as True to sort in descending order.

        Raises:
            NotImplementedError: The sort method for segment init from client is not supported yet.

        """
        try:
            self._data.sort(key=key, reverse=reverse)  # type: ignore[attr-defined]
        except AttributeError as error:
            raise NotImplementedError(
                "The sort method for segment init from client is not supported yet."
            ) from error


class FusionSegment(NameMixin, UserMutableSequence[Frame]):
    """This class defines the concept of fusion segment.

    Fusion segment is a concept in :class:`~tensorbay.dataset.dataset.FusionDataset`.
    It is the structure that composes :class:`~tensorbay.dataset.dataset.FusionDataset`,
    and consists of a list of :class:`~tensorbay.dataset.frame.Frame`.

    Besides, a fusion segment contains multiple :class:`~tensorbay.sensor.sensor.Sensors`
    correspoinding to the :class:`~tensorbay.dataset.data.Data`
    under each :class:`~tensorbay.dataset.frame.Frame`.

    If the segment is inside of a time-continuous :class:`~tensorbay.dataset.dataset.FusionDataset`,
    the time continuity of the frames should be indicated by the index inside the fusion segment.

    Since :class:`FusionSegment` extends :class:`~tensorbay.utility.user.UserMutableSequence`,
    its basic operations are the same as a list's.

    To initialize a :class:`FusionSegment` and add a
    :class:`~tensorbay.dataset.frame.Frame` to it:

    .. code:: python

        fusion_segment = FusionSegment(fusion_segment_name)
        frame = Frame()
        ...
        fusion_segment.append(frame)

    Arguments:
        name: The name of the fusion segment, whose default value is an empty string.
        client: The FusionDatasetClient if you want to read the segment from tensorbay.

    """

    _repr_type = ReprType.SEQUENCE
    _repr_attrs = ("sensors",)
    _repr_maxlevel = 2

    def __init__(
        self, name: str = "default", client: Optional["FusionDatasetClient"] = None
    ) -> None:
        super().__init__(name)

        self._data: MutableSequence[Frame]
        if client:
            self._client = client.get_segment(name)
            self._data = self._client.list_frames()
            self._repr_non_empty = True
        else:
            self._data = []
            self._sensors = Sensors()

    @classmethod
    def _from_client(cls: Type[_FS], client: "FusionSegmentClient") -> _FS:
        """Init a FusionSegment from :class:`~tensorbay.client.segment.FusionSegmentClient`.

        Arguments:
            client: The :class:`~tensorbay.client.segment.FusionSegmentClient`.

        Returns:
            The FusionSegment of the input :class:`~tensorbay.client.segment.FusionSegmentClient`.

        """
        segment: _FS = object.__new__(cls)
        super(cls, segment).__init__(name=client.name)  # pylint: disable=no-value-for-parameter
        segment._client = client
        segment._data = client.list_frames()
        segment._repr_non_empty = True
        return segment

    @property
    def sensors(self) -> Sensors:
        """Return the sensors of the fusion segment.

        Returns:
            The :class:`~tensorbay.sensor.sensor.Sensors` of the fusion dataset.

        """
        if not hasattr(self, "_sensors"):
            self._sensors = self._client.get_sensors()

        return self._sensors

    @sensors.setter
    def sensors(self, sensors: Sensors) -> None:
        self._sensors = sensors
